from invokeai.app.invocations.baseinvocation import (
    BaseInvocation,
    BaseInvocationOutput,
    invocation,
    invocation_output,
)
from invokeai.app.invocations.fields import (
    FieldDescriptions,
    FluxKontextConditioningField,
    InputField,
    OutputField,
)
from invokeai.app.invocations.primitives import ImageField
from invokeai.app.services.shared.invocation_context import InvocationContext


@invocation_output("flux_kontext_output")
class FluxKontextOutput(BaseInvocationOutput):
    """The conditioning output of a FLUX Kontext invocation."""

    kontext_cond: FluxKontextConditioningField = OutputField(
        description=FieldDescriptions.flux_kontext_conditioning, title="Kontext Conditioning"
    )


@invocation(
    "flux_kontext",
    title="Kontext Conditioning - FLUX",
    tags=["conditioning", "kontext", "flux"],
    category="conditioning",
    version="1.0.0",
)
class FluxKontextInvocation(BaseInvocation):
    """Prepares a reference image for FLUX Kontext conditioning."""

    image: ImageField = InputField(description="The Kontext reference image.")

    def invoke(self, context: InvocationContext) -> FluxKontextOutput:
        """Packages the provided image into a Kontext conditioning field."""
        return FluxKontextOutput(kontext_cond=FluxKontextConditioningField(image=self.image))



from fastapi import Body, HTTPException
from fastapi.routing import APIRouter

from invokeai.app.services.videos_common import AddVideosToBoardResult, RemoveVideosFromBoardResult

board_videos_router = APIRouter(prefix="/v1/board_videos", tags=["boards"])


@board_videos_router.post(
    "/batch",
    operation_id="add_videos_to_board",
    responses={
        201: {"description": "Videos were added to board successfully"},
    },
    status_code=201,
    response_model=AddVideosToBoardResult,
)
async def add_videos_to_board(
    board_id: str = Body(description="The id of the board to add to"),
    video_ids: list[str] = Body(description="The ids of the videos to add", embed=True),
) -> AddVideosToBoardResult:
    """Adds a list of videos to a board"""
    raise HTTPException(status_code=501, detail="Not implemented")


@board_videos_router.post(
    "/batch/delete",
    operation_id="remove_videos_from_board",
    responses={
        201: {"description": "Videos were removed from board successfully"},
    },
    status_code=201,
    response_model=RemoveVideosFromBoardResult,
)
async def remove_videos_from_board(
    video_ids: list[str] = Body(description="The ids of the videos to remove", embed=True),
) -> RemoveVideosFromBoardResult:
    """Removes a list of videos from their board, if they had one"""
    raise HTTPException(status_code=501, detail="Not implemented")



"""Utility functions for extracting metadata from LoRA model files."""

import json
import logging
from pathlib import Path
from typing import Any, Dict, Optional, Set, Tuple

from PIL import Image

from invokeai.app.util.thumbnails import make_thumbnail
from invokeai.backend.model_manager.configs.factory import AnyModelConfig
from invokeai.backend.model_manager.taxonomy import ModelType

logger = logging.getLogger(__name__)


def extract_lora_metadata(
    model_path: Path, model_key: str, model_images_path: Path
) -> Tuple[Optional[str], Optional[Set[str]]]:
    """
    Extract metadata for a LoRA model from associated JSON and image files.

    Args:
        model_path: Path to the LoRA model file
        model_key: Unique key for the model
        model_images_path: Path to the model images directory

    Returns:
        Tuple of (description, trigger_phrases)
    """
    model_stem = model_path.stem
    model_dir = model_path.parent

    # Find and process preview image
    _process_preview_image(model_stem, model_dir, model_key, model_images_path)

    # Extract metadata from JSON
    description, trigger_phrases = _extract_json_metadata(model_stem, model_dir)

    return description, trigger_phrases


def _process_preview_image(model_stem: str, model_dir: Path, model_key: str, model_images_path: Path) -> bool:
    """Find and process a preview image for the model, saving it to the model images store."""
    image_extensions = [".png", ".jpg", ".jpeg", ".webp"]

    for ext in image_extensions:
        image_path = model_dir / f"{model_stem}{ext}"
        if image_path.exists():
            try:
                # Open the image
                with Image.open(image_path) as img:
                    # Create thumbnail and save to model images directory
                    thumbnail = make_thumbnail(img, 256)
                    thumbnail_path = model_images_path / f"{model_key}.webp"
                    thumbnail.save(thumbnail_path, format="webp")

                logger.info(f"Processed preview image {image_path.name} for model {model_key}")
                return True

            except Exception as e:
                logger.warning(f"Failed to process preview image {image_path.name}: {e}")
                return False

    return False


def _extract_json_metadata(model_stem: str, model_dir: Path) -> Tuple[Optional[str], Optional[Set[str]]]:
    """Extract metadata from a JSON file with the same name as the model."""
    json_path = model_dir / f"{model_stem}.json"

    if not json_path.exists():
        return None, None

    try:
        with open(json_path, "r", encoding="utf-8") as f:
            metadata = json.load(f)

        # Extract description
        description = _build_description(metadata)

        # Extract trigger phrases
        trigger_phrases = _extract_trigger_phrases(metadata)

        if description or trigger_phrases:
            logger.info(f"Applied metadata from {json_path.name}")

        return description, trigger_phrases

    except (json.JSONDecodeError, IOError, Exception) as e:
        logger.warning(f"Failed to read metadata from {json_path}: {e}")
        return None, None


def _build_description(metadata: Dict[str, Any]) -> Optional[str]:
    """Build a description from metadata fields."""
    description_parts = []

    if description := metadata.get("description"):
        description_parts.append(str(description).strip())

    if notes := metadata.get("notes"):
        description_parts.append(str(notes).strip())

    return " | ".join(description_parts) if description_parts else None


def _extract_trigger_phrases(metadata: Dict[str, Any]) -> Optional[Set[str]]:
    """Extract trigger phrases from metadata."""
    if not (activation_text := metadata.get("activation text")):
        return None

    activation_text = str(activation_text).strip()
    if not activation_text:
        return None

    # Split on commas and clean up each phrase
    phrases = [phrase.strip() for phrase in activation_text.split(",") if phrase.strip()]

    return set(phrases) if phrases else None


def apply_lora_metadata(info: AnyModelConfig, model_path: Path, model_images_path: Path) -> None:
    """
    Apply extracted metadata to a LoRA model configuration.

    Args:
        info: The model configuration to update
        model_path: Path to the LoRA model file
        model_images_path: Path to the model images directory
    """
    # Only process LoRA models
    if info.type != ModelType.LoRA:
        return

    # Extract and apply metadata
    description, trigger_phrases = extract_lora_metadata(model_path, info.key, model_images_path)

    # We don't set cover_image path in the config anymore since images are stored
    # separately in the model images store by model key

    if description:
        info.description = description

    if trigger_phrases:
        info.trigger_phrases = trigger_phrases



from invokeai.backend.model_manager.omi.vendor.convert.lora.convert_lora_util import (
    LoraConversionKeySet,
    map_prefix_range,
)


def map_clip(key_prefix: LoraConversionKeySet) -> list[LoraConversionKeySet]:
    keys = []

    keys += [LoraConversionKeySet("text_projection", "text_projection", parent=key_prefix)]

    for k in map_prefix_range("text_model.encoder.layers", "text_model.encoder.layers", parent=key_prefix):
        keys += [LoraConversionKeySet("mlp.fc1", "mlp.fc1", parent=k)]
        keys += [LoraConversionKeySet("mlp.fc2", "mlp.fc2", parent=k)]
        keys += [LoraConversionKeySet("self_attn.k_proj", "self_attn.k_proj", parent=k)]
        keys += [LoraConversionKeySet("self_attn.out_proj", "self_attn.out_proj", parent=k)]
        keys += [LoraConversionKeySet("self_attn.q_proj", "self_attn.q_proj", parent=k)]
        keys += [LoraConversionKeySet("self_attn.v_proj", "self_attn.v_proj", parent=k)]

    return keys



import sqlite3
from logging import Logger

from invokeai.app.services.config import InvokeAIAppConfig
from invokeai.app.services.shared.sqlite_migrator.sqlite_migrator_common import Migration


class Migration22Callback:
    def __init__(self, app_config: InvokeAIAppConfig, logger: Logger) -> None:
        self._app_config = app_config
        self._logger = logger
        self._models_dir = app_config.models_path.resolve()

    def __call__(self, cursor: sqlite3.Cursor) -> None:
        self._logger.info("Removing UNIQUE(name, base, type) constraint from models table")

        # Step 1: Rename the existing models table
        cursor.execute("ALTER TABLE models RENAME TO models_old;")

        # Step 2: Create the new models table without the UNIQUE(name, base, type) constraint
        cursor.execute(
            """--sql
            CREATE TABLE models (
                id TEXT NOT NULL PRIMARY KEY,
                hash TEXT GENERATED ALWAYS as (json_extract(config, '$.hash')) VIRTUAL NOT NULL,
                base TEXT GENERATED ALWAYS as (json_extract(config, '$.base')) VIRTUAL NOT NULL,
                type TEXT GENERATED ALWAYS as (json_extract(config, '$.type')) VIRTUAL NOT NULL,
                path TEXT GENERATED ALWAYS as (json_extract(config, '$.path')) VIRTUAL NOT NULL,
                format TEXT GENERATED ALWAYS as (json_extract(config, '$.format')) VIRTUAL NOT NULL,
                name TEXT GENERATED ALWAYS as (json_extract(config, '$.name')) VIRTUAL NOT NULL,
                description TEXT GENERATED ALWAYS as (json_extract(config, '$.description')) VIRTUAL,
                source TEXT GENERATED ALWAYS as (json_extract(config, '$.source')) VIRTUAL NOT NULL,
                source_type TEXT GENERATED ALWAYS as (json_extract(config, '$.source_type')) VIRTUAL NOT NULL,
                source_api_response TEXT GENERATED ALWAYS as (json_extract(config, '$.source_api_response')) VIRTUAL,
                trigger_phrases TEXT GENERATED ALWAYS as (json_extract(config, '$.trigger_phrases')) VIRTUAL,
                file_size INTEGER GENERATED ALWAYS as (json_extract(config, '$.file_size')) VIRTUAL NOT NULL,
                -- Serialized JSON representation of the whole config object, which will contain additional fields from subclasses
                config TEXT NOT NULL,
                created_at DATETIME NOT NULL DEFAULT(STRFTIME('%Y-%m-%d %H:%M:%f', 'NOW')),
                -- Updated via trigger
                updated_at DATETIME NOT NULL DEFAULT(STRFTIME('%Y-%m-%d %H:%M:%f', 'NOW')),
                -- Explicit unique constraint on path
                UNIQUE(path)
            );
            """
        )

        # Step 3: Copy all data from the old table to the new table
        # Only copy the stored columns (id, config, created_at, updated_at), not the virtual columns
        cursor.execute(
            "INSERT INTO models (id, config, created_at, updated_at) "
            "SELECT id, config, created_at, updated_at FROM models_old;"
        )

        # Step 4: Drop the old table
        cursor.execute("DROP TABLE models_old;")

        # Step 5: Recreate indexes
        cursor.execute("CREATE INDEX IF NOT EXISTS base_index ON models(base);")
        cursor.execute("CREATE INDEX IF NOT EXISTS type_index ON models(type);")
        cursor.execute("CREATE INDEX IF NOT EXISTS name_index ON models(name);")

        # Step 6: Recreate the updated_at trigger
        cursor.execute(
            """--sql
            CREATE TRIGGER models_updated_at
            AFTER UPDATE
            ON models FOR EACH ROW
            BEGIN
                UPDATE models SET updated_at = STRFTIME('%Y-%m-%d %H:%M:%f', 'NOW')
                    WHERE id = old.id;
            END;
            """
        )


def build_migration_22(app_config: InvokeAIAppConfig, logger: Logger) -> Migration:
    """Builds the migration object for migrating from version 21 to version 22.

    This migration:
    - Removes the UNIQUE constraint on the combination of (base, name, type) columns in the models table
    - Adds an explicit UNIQUE contraint on the path column
    """

    return Migration(
        from_version=21,
        to_version=22,
        callback=Migration22Callback(app_config=app_config, logger=logger),
    )



from typing import (
    Literal,
    Self,
)

from pydantic import Field
from typing_extensions import Any

from invokeai.backend.model_manager.configs.base import Config_Base, Diffusers_Config_Base
from invokeai.backend.model_manager.configs.identification_utils import (
    NotAMatchError,
    get_class_name_from_config_dict_or_raise,
    get_config_dict_or_raise,
    raise_for_override_fields,
    raise_if_not_dir,
)
from invokeai.backend.model_manager.model_on_disk import ModelOnDisk
from invokeai.backend.model_manager.taxonomy import (
    BaseModelType,
    ModelFormat,
    ModelType,
)


class CLIPVision_Diffusers_Config(Diffusers_Config_Base, Config_Base):
    """Model config for CLIPVision."""

    base: Literal[BaseModelType.Any] = Field(default=BaseModelType.Any)
    type: Literal[ModelType.CLIPVision] = Field(default=ModelType.CLIPVision)
    format: Literal[ModelFormat.Diffusers] = Field(default=ModelFormat.Diffusers)

    @classmethod
    def from_model_on_disk(cls, mod: ModelOnDisk, override_fields: dict[str, Any]) -> Self:
        raise_if_not_dir(mod)

        raise_for_override_fields(cls, override_fields)

        cls.raise_if_config_doesnt_look_like_clip_vision(mod)

        return cls(**override_fields)

    @classmethod
    def raise_if_config_doesnt_look_like_clip_vision(cls, mod: ModelOnDisk) -> None:
        config_dict = get_config_dict_or_raise(mod.path / "config.json")
        class_name = get_class_name_from_config_dict_or_raise(config_dict)

        if class_name == "CLIPVisionModelWithProjection":
            looks_like_clip_vision = True
        elif class_name == "CLIPModel" and "vision_config" in config_dict:
            looks_like_clip_vision = True
        else:
            looks_like_clip_vision = False

        if not looks_like_clip_vision:
            raise NotAMatchError(
                f"config class name is {class_name}, not CLIPVisionModelWithProjection or CLIPModel with vision_config"
            )



# This file is vendored from https://github.com/ShieldMnt/invisible-watermark
#
# `invisible-watermark` is MIT licensed as of August 23, 2025, when the code was copied into this repo.
#
# Why we vendored it in:
# `invisible-watermark` has a dependency on `opencv-python`, which conflicts with Invoke's dependency on
# `opencv-contrib-python`. It's easier to copy the code over than complicate the installation process by
# requiring an extra post-install step of removing `opencv-python` and installing `opencv-contrib-python`.

import struct
import uuid
import base64
import cv2
import numpy as np
import pywt


class WatermarkEncoder(object):
    def __init__(self, content=b""):
        seq = np.array([n for n in content], dtype=np.uint8)
        self._watermarks = list(np.unpackbits(seq))
        self._wmLen = len(self._watermarks)
        self._wmType = "bytes"

    def set_by_ipv4(self, addr):
        bits = []
        ips = addr.split(".")
        for ip in ips:
            bits += list(np.unpackbits(np.array([ip % 255], dtype=np.uint8)))
        self._watermarks = bits
        self._wmLen = len(self._watermarks)
        self._wmType = "ipv4"
        assert self._wmLen == 32

    def set_by_uuid(self, uid):
        u = uuid.UUID(uid)
        self._wmType = "uuid"
        seq = np.array([n for n in u.bytes], dtype=np.uint8)
        self._watermarks = list(np.unpackbits(seq))
        self._wmLen = len(self._watermarks)

    def set_by_bytes(self, content):
        self._wmType = "bytes"
        seq = np.array([n for n in content], dtype=np.uint8)
        self._watermarks = list(np.unpackbits(seq))
        self._wmLen = len(self._watermarks)

    def set_by_b16(self, b16):
        content = base64.b16decode(b16)
        self.set_by_bytes(content)
        self._wmType = "b16"

    def set_by_bits(self, bits=[]):
        self._watermarks = [int(bit) % 2 for bit in bits]
        self._wmLen = len(self._watermarks)
        self._wmType = "bits"

    def set_watermark(self, wmType="bytes", content=""):
        if wmType == "ipv4":
            self.set_by_ipv4(content)
        elif wmType == "uuid":
            self.set_by_uuid(content)
        elif wmType == "bits":
            self.set_by_bits(content)
        elif wmType == "bytes":
            self.set_by_bytes(content)
        elif wmType == "b16":
            self.set_by_b16(content)
        else:
            raise NameError("%s is not supported" % wmType)

    def get_length(self):
        return self._wmLen

    # @classmethod
    # def loadModel(cls):
    #     RivaWatermark.loadModel()

    def encode(self, cv2Image, method="dwtDct", **configs):
        (r, c, channels) = cv2Image.shape
        if r * c < 256 * 256:
            raise RuntimeError("image too small, should be larger than 256x256")

        if method == "dwtDct":
            embed = EmbedMaxDct(self._watermarks, wmLen=self._wmLen, **configs)
            return embed.encode(cv2Image)
        # elif method == 'dwtDctSvd':
        #     embed = EmbedDwtDctSvd(self._watermarks, wmLen=self._wmLen, **configs)
        #     return embed.encode(cv2Image)
        # elif method == 'rivaGan':
        #     embed = RivaWatermark(self._watermarks, self._wmLen)
        #     return embed.encode(cv2Image)
        else:
            raise NameError("%s is not supported" % method)


class WatermarkDecoder(object):
    def __init__(self, wm_type="bytes", length=0):
        self._wmType = wm_type
        if wm_type == "ipv4":
            self._wmLen = 32
        elif wm_type == "uuid":
            self._wmLen = 128
        elif wm_type == "bytes":
            self._wmLen = length
        elif wm_type == "bits":
            self._wmLen = length
        elif wm_type == "b16":
            self._wmLen = length
        else:
            raise NameError("%s is unsupported" % wm_type)

    def reconstruct_ipv4(self, bits):
        ips = [str(ip) for ip in list(np.packbits(bits))]
        return ".".join(ips)

    def reconstruct_uuid(self, bits):
        nums = np.packbits(bits)
        bstr = b""
        for i in range(16):
            bstr += struct.pack(">B", nums[i])

        return str(uuid.UUID(bytes=bstr))

    def reconstruct_bits(self, bits):
        # return ''.join([str(b) for b in bits])
        return bits

    def reconstruct_b16(self, bits):
        bstr = self.reconstruct_bytes(bits)
        return base64.b16encode(bstr)

    def reconstruct_bytes(self, bits):
        nums = np.packbits(bits)
        bstr = b""
        for i in range(self._wmLen // 8):
            bstr += struct.pack(">B", nums[i])
        return bstr

    def reconstruct(self, bits):
        if len(bits) != self._wmLen:
            raise RuntimeError("bits are not matched with watermark length")

        if self._wmType == "ipv4":
            return self.reconstruct_ipv4(bits)
        elif self._wmType == "uuid":
            return self.reconstruct_uuid(bits)
        elif self._wmType == "bits":
            return self.reconstruct_bits(bits)
        elif self._wmType == "b16":
            return self.reconstruct_b16(bits)
        else:
            return self.reconstruct_bytes(bits)

    def decode(self, cv2Image, method="dwtDct", **configs):
        (r, c, channels) = cv2Image.shape
        if r * c < 256 * 256:
            raise RuntimeError("image too small, should be larger than 256x256")

        bits = []
        if method == "dwtDct":
            embed = EmbedMaxDct(watermarks=[], wmLen=self._wmLen, **configs)
            bits = embed.decode(cv2Image)
        # elif method == 'dwtDctSvd':
        #     embed = EmbedDwtDctSvd(watermarks=[], wmLen=self._wmLen, **configs)
        #     bits = embed.decode(cv2Image)
        # elif method == 'rivaGan':
        #     embed = RivaWatermark(watermarks=[], wmLen=self._wmLen, **configs)
        #     bits = embed.decode(cv2Image)
        else:
            raise NameError("%s is not supported" % method)
        return self.reconstruct(bits)

    # @classmethod
    # def loadModel(cls):
    #     RivaWatermark.loadModel()


class EmbedMaxDct(object):
    def __init__(self, watermarks=[], wmLen=8, scales=[0, 36, 36], block=4):
        self._watermarks = watermarks
        self._wmLen = wmLen
        self._scales = scales
        self._block = block

    def encode(self, bgr):
        (row, col, channels) = bgr.shape

        yuv = cv2.cvtColor(bgr, cv2.COLOR_BGR2YUV)

        for channel in range(2):
            if self._scales[channel] <= 0:
                continue

            ca1, (h1, v1, d1) = pywt.dwt2(yuv[: row // 4 * 4, : col // 4 * 4, channel], "haar")
            self.encode_frame(ca1, self._scales[channel])

            yuv[: row // 4 * 4, : col // 4 * 4, channel] = pywt.idwt2((ca1, (v1, h1, d1)), "haar")

        bgr_encoded = cv2.cvtColor(yuv, cv2.COLOR_YUV2BGR)
        return bgr_encoded

    def decode(self, bgr):
        (row, col, channels) = bgr.shape

        yuv = cv2.cvtColor(bgr, cv2.COLOR_BGR2YUV)

        scores = [[] for i in range(self._wmLen)]
        for channel in range(2):
            if self._scales[channel] <= 0:
                continue

            ca1, (h1, v1, d1) = pywt.dwt2(yuv[: row // 4 * 4, : col // 4 * 4, channel], "haar")

            scores = self.decode_frame(ca1, self._scales[channel], scores)

        avgScores = list(map(lambda l: np.array(l).mean(), scores))

        bits = np.array(avgScores) * 255 > 127
        return bits

    def decode_frame(self, frame, scale, scores):
        (row, col) = frame.shape
        num = 0

        for i in range(row // self._block):
            for j in range(col // self._block):
                block = frame[
                    i * self._block : i * self._block + self._block, j * self._block : j * self._block + self._block
                ]

                score = self.infer_dct_matrix(block, scale)
                # score = self.infer_dct_svd(block, scale)
                wmBit = num % self._wmLen
                scores[wmBit].append(score)
                num = num + 1

        return scores

    def diffuse_dct_svd(self, block, wmBit, scale):
        u, s, v = np.linalg.svd(cv2.dct(block))

        s[0] = (s[0] // scale + 0.25 + 0.5 * wmBit) * scale
        return cv2.idct(np.dot(u, np.dot(np.diag(s), v)))

    def infer_dct_svd(self, block, scale):
        u, s, v = np.linalg.svd(cv2.dct(block))

        score = 0
        score = int((s[0] % scale) > scale * 0.5)
        return score
        if score >= 0.5:
            return 1.0
        else:
            return 0.0

    def diffuse_dct_matrix(self, block, wmBit, scale):
        pos = np.argmax(abs(block.flatten()[1:])) + 1
        i, j = pos // self._block, pos % self._block
        val = block[i][j]
        if val >= 0.0:
            block[i][j] = (val // scale + 0.25 + 0.5 * wmBit) * scale
        else:
            val = abs(val)
            block[i][j] = -1.0 * (val // scale + 0.25 + 0.5 * wmBit) * scale
        return block

    def infer_dct_matrix(self, block, scale):
        pos = np.argmax(abs(block.flatten()[1:])) + 1
        i, j = pos // self._block, pos % self._block

        val = block[i][j]
        if val < 0:
            val = abs(val)

        if (val % scale) > 0.5 * scale:
            return 1
        else:
            return 0

    def encode_frame(self, frame, scale):
        """
        frame is a matrix (M, N)

        we get K (watermark bits size) blocks (self._block x self._block)

        For i-th block, we encode watermark[i] bit into it
        """
        (row, col) = frame.shape
        num = 0
        for i in range(row // self._block):
            for j in range(col // self._block):
                block = frame[
                    i * self._block : i * self._block + self._block, j * self._block : j * self._block + self._block
                ]
                wmBit = self._watermarks[(num % self._wmLen)]

                diffusedBlock = self.diffuse_dct_matrix(block, wmBit, scale)
                # diffusedBlock = self.diffuse_dct_svd(block, wmBit, scale)
                frame[
                    i * self._block : i * self._block + self._block, j * self._block : j * self._block + self._block
                ] = diffusedBlock

                num = num + 1



from abc import ABC, abstractmethod
from enum import Enum
from inspect import isabstract
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Literal,
    Self,
    Type,
)

from pydantic import BaseModel, ConfigDict, Field, Tag
from pydantic_core import PydanticUndefined

from invokeai.app.util.misc import uuid_string
from invokeai.backend.model_manager.model_on_disk import ModelOnDisk
from invokeai.backend.model_manager.taxonomy import (
    AnyVariant,
    BaseModelType,
    ModelFormat,
    ModelRepoVariant,
    ModelSourceType,
    ModelType,
)

if TYPE_CHECKING:
    pass


class Config_Base(ABC, BaseModel):
    """
    Abstract base class for model configurations. A model config describes a specific combination of model base, type and
    format, along with other metadata about the model. For example, a Stable Diffusion 1.x main model in checkpoint format
    would have base=sd-1, type=main, format=checkpoint.

    To create a new config type, inherit from this class and implement its interface:
    - Define method 'from_model_on_disk' that returns an instance of the class or raises NotAMatch. This method will be
        called during model installation to determine the correct config class for a model.
    - Define fields 'type', 'base' and 'format' as pydantic fields. These should be Literals with a single value. A
        default must be provided for each of these fields.

    If multiple combinations of base, type and format need to be supported, create a separate subclass for each.

    See MinimalConfigExample in test_model_probe.py for an example implementation.
    """

    # These fields are common to all model configs.

    key: str = Field(
        default_factory=uuid_string,
        description="A unique key for this model.",
    )
    hash: str = Field(
        description="The hash of the model file(s).",
    )
    path: str = Field(
        description="Path to the model on the filesystem. Relative paths are relative to the Invoke root directory.",
    )
    file_size: int = Field(
        description="The size of the model in bytes.",
    )
    name: str = Field(
        description="Name of the model.",
    )
    description: str | None = Field(
        default=None,
        description="Model description",
    )
    source: str = Field(
        description="The original source of the model (path, URL or repo_id).",
    )
    source_type: ModelSourceType = Field(
        description="The type of source",
    )
    source_api_response: str | None = Field(
        default=None,
        description="The original API response from the source, as stringified JSON.",
    )
    cover_image: str | None = Field(
        default=None,
        description="Url for image to preview model",
    )
    usage_info: str | None = Field(
        default=None,
        description="Usage information for this model",
    )

    CONFIG_CLASSES: ClassVar[set[Type["Config_Base"]]] = set()
    """Set of all non-abstract subclasses of Config_Base, for use during model probing. In other words, this is the set
    of all known model config types."""

    model_config = ConfigDict(
        validate_assignment=True,
        json_schema_serialization_defaults_required=True,
        json_schema_mode_override="serialization",
    )

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        # Register non-abstract subclasses so we can iterate over them later during model probing. Note that
        # isabstract() will return False if the class does not have any abstract methods, even if it inherits from ABC.
        # We must check for ABC lest we unintentionally register some abstract model config classes.
        if not isabstract(cls) and ABC not in cls.__bases__:
            cls.CONFIG_CLASSES.add(cls)

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs):
        # Ensure that model configs define 'base', 'type' and 'format' fields and provide defaults for them. Each
        # subclass is expected to represent a single combination of base, type and format.
        #
        # This pydantic dunder method is called after the pydantic model for a class is created. The normal
        # __init_subclass__ is too early to do this check.
        for name in ("type", "base", "format"):
            if name not in cls.model_fields:
                raise NotImplementedError(f"{cls.__name__} must define a '{name}' field")
            if cls.model_fields[name].default is PydanticUndefined:
                raise NotImplementedError(f"{cls.__name__} must define a default for the '{name}' field")

    @classmethod
    def get_tag(cls) -> Tag:
        """Constructs a pydantic discriminated union tag for this model config class. When a config is deserialized,
        pydantic uses the tag to determine which subclass to instantiate.

        The tag is a dot-separated string of the type, format, base and variant (if applicable).
        """
        tag_strings: list[str] = []
        for name in ("type", "format", "base", "variant"):
            if field := cls.model_fields.get(name):
                # The check in __pydantic_init_subclass__ ensures that type, format and base are always present with
                # defaults. variant does not require a default, but if it has one, we need to add it to the tag. We can
                # check for the presence of a default by seeing if it's not PydanticUndefined, a sentinel value used by
                # pydantic to indicate that no default was provided.
                if field.default is not PydanticUndefined:
                    # We expect each of these fields has an Enum for its default; we want the value of the enum.
                    tag_strings.append(field.default.value)
        return Tag(".".join(tag_strings))

    @staticmethod
    def get_model_discriminator_value(v: Any) -> str:
        """Computes the discriminator value for a model config discriminated union."""
        # This is called by pydantic during deserialization and serialization to determine which model the data
        # represents. It can get either a dict (during deserialization) or an instance of a Config_Base subclass
        # (during serialization).
        #
        # See: https://docs.pydantic.dev/latest/concepts/unions/#discriminated-unions-with-callable-discriminator
        if isinstance(v, Config_Base):
            # We have an instance of a ModelConfigBase subclass - use its tag directly.
            return v.get_tag().tag
        if isinstance(v, dict):
            # We have a dict - attempt to compute a tag from its fields.
            tag_strings: list[str] = []
            if type_ := v.get("type"):
                if isinstance(type_, Enum):
                    type_ = str(type_.value)
                elif not isinstance(type_, str):
                    raise ValueError("Model config dict 'type' field must be a string or Enum")
                tag_strings.append(type_)

            if format_ := v.get("format"):
                if isinstance(format_, Enum):
                    format_ = str(format_.value)
                elif not isinstance(format_, str):
                    raise ValueError("Model config dict 'format' field must be a string or Enum")
                tag_strings.append(format_)

            if base_ := v.get("base"):
                if isinstance(base_, Enum):
                    base_ = str(base_.value)
                elif not isinstance(base_, str):
                    raise ValueError("Model config dict 'base' field must be a string or Enum")
                tag_strings.append(base_)

            # Special case: CLIP Embed models also need the variant to distinguish them.
            if (
                type_ == ModelType.CLIPEmbed.value
                and format_ == ModelFormat.Diffusers.value
                and base_ == BaseModelType.Any.value
            ):
                if variant_ := v.get("variant"):
                    if isinstance(variant_, Enum):
                        variant_ = variant_.value
                    elif not isinstance(variant_, str):
                        raise ValueError("Model config dict 'variant' field must be a string or Enum")
                    tag_strings.append(variant_)
                else:
                    raise ValueError("CLIP Embed model config dict must include a 'variant' field")

            return ".".join(tag_strings)
        else:
            raise ValueError(
                "Model config discriminator value must be computed from a dict or ModelConfigBase instance"
            )

    @classmethod
    @abstractmethod
    def from_model_on_disk(cls, mod: ModelOnDisk, override_fields: dict[str, Any]) -> Self:
        """Given the model on disk and any override fields, attempt to construct an instance of this config class.

        This method serves to identify whether the model on disk matches this config class, and if so, to extract any
        additional metadata needed to instantiate the config.

        Implementations should raise a NotAMatchError if the model does not match this config class."""
        raise NotImplementedError(f"from_model_on_disk not implemented for {cls.__name__}")


class Checkpoint_Config_Base(ABC, BaseModel):
    """Base class for checkpoint-style models."""

    config_path: str | None = Field(
        description="Path to the config for this model, if any.",
        default=None,
    )


class Diffusers_Config_Base(ABC, BaseModel):
    """Base class for diffusers-style models."""

    format: Literal[ModelFormat.Diffusers] = Field(default=ModelFormat.Diffusers)
    repo_variant: ModelRepoVariant = Field(ModelRepoVariant.Default)

    @classmethod
    def _get_repo_variant_or_raise(cls, mod: ModelOnDisk) -> ModelRepoVariant:
        # get all files ending in .bin or .safetensors
        weight_files = list(mod.path.glob("**/*.safetensors"))
        weight_files.extend(list(mod.path.glob("**/*.bin")))
        for x in weight_files:
            if ".fp16" in x.suffixes:
                return ModelRepoVariant.FP16
            if "openvino_model" in x.name:
                return ModelRepoVariant.OpenVINO
            if "flax_model" in x.name:
                return ModelRepoVariant.Flax
            if x.suffix == ".onnx":
                return ModelRepoVariant.ONNX
        return ModelRepoVariant.Default


class SubmodelDefinition(BaseModel):
    path_or_prefix: str
    model_type: ModelType
    variant: AnyVariant | None = None

    model_config = ConfigDict(protected_namespaces=())



from invokeai.backend.model_manager.model_on_disk import StateDict
from invokeai.backend.model_manager.omi.vendor.convert.lora import (
    convert_flux_lora as omi_flux,
)
from invokeai.backend.model_manager.omi.vendor.convert.lora import (
    convert_lora_util as lora_util,
)
from invokeai.backend.model_manager.omi.vendor.convert.lora import (
    convert_sdxl_lora as omi_sdxl,
)
from invokeai.backend.model_manager.taxonomy import BaseModelType


def convert_from_omi(weights_sd: StateDict, base: BaseModelType):
    keyset = {
        BaseModelType.Flux: omi_flux.convert_flux_lora_key_sets(),
        BaseModelType.StableDiffusionXL: omi_sdxl.convert_sdxl_lora_key_sets(),
    }[base]
    source = "omi"
    target = "legacy_diffusers"
    return lora_util.__convert(weights_sd, keyset, source, target)  # type: ignore



from abc import ABC
from pathlib import Path
from typing import (
    Literal,
    Self,
)

import torch
from pydantic import BaseModel, Field
from typing_extensions import Any

from invokeai.backend.model_manager.configs.base import Config_Base
from invokeai.backend.model_manager.configs.identification_utils import (
    NotAMatchError,
    raise_for_override_fields,
    raise_if_not_dir,
    raise_if_not_file,
)
from invokeai.backend.model_manager.model_on_disk import ModelOnDisk
from invokeai.backend.model_manager.taxonomy import (
    BaseModelType,
    ModelFormat,
    ModelType,
)


class TI_Config_Base(ABC, BaseModel):
    type: Literal[ModelType.TextualInversion] = Field(default=ModelType.TextualInversion)

    @classmethod
    def _validate_base(cls, mod: ModelOnDisk, path: Path | None = None) -> None:
        expected_base = cls.model_fields["base"].default
        recognized_base = cls._get_base_or_raise(mod, path)
        if expected_base is not recognized_base:
            raise NotAMatchError(f"base is {recognized_base}, not {expected_base}")

    @classmethod
    def _file_looks_like_embedding(cls, mod: ModelOnDisk, path: Path | None = None) -> bool:
        try:
            p = path or mod.path

            if not p.exists():
                return False

            if p.is_dir():
                return False

            if p.name in [f"learned_embeds.{s}" for s in mod.weight_files()]:
                return True

            state_dict = mod.load_state_dict(p)

            # Heuristic: textual inversion embeddings have these keys
            if any(key in {"string_to_param", "emb_params", "clip_g"} for key in state_dict.keys()):
                return True

            # Heuristic: small state dict with all tensor values
            if (len(state_dict)) < 10 and all(isinstance(v, torch.Tensor) for v in state_dict.values()):
                return True

            return False
        except Exception:
            return False

    @classmethod
    def _get_base_or_raise(cls, mod: ModelOnDisk, path: Path | None = None) -> BaseModelType:
        p = path or mod.path

        try:
            state_dict = mod.load_state_dict(p)
        except Exception as e:
            raise NotAMatchError(f"unable to load state dict from {p}: {e}") from e

        try:
            if "string_to_token" in state_dict:
                token_dim = list(state_dict["string_to_param"].values())[0].shape[-1]
            elif "emb_params" in state_dict:
                token_dim = state_dict["emb_params"].shape[-1]
            elif "clip_g" in state_dict:
                token_dim = state_dict["clip_g"].shape[-1]
            else:
                token_dim = list(state_dict.values())[0].shape[0]
        except Exception as e:
            raise NotAMatchError(f"unable to determine token dimension from state dict in {p}: {e}") from e

        match token_dim:
            case 768:
                return BaseModelType.StableDiffusion1
            case 1024:
                return BaseModelType.StableDiffusion2
            case 1280:
                return BaseModelType.StableDiffusionXL
            case _:
                raise NotAMatchError(f"unrecognized token dimension {token_dim}")


class TI_File_Config_Base(TI_Config_Base):
    """Model config for textual inversion embeddings."""

    format: Literal[ModelFormat.EmbeddingFile] = Field(default=ModelFormat.EmbeddingFile)

    @classmethod
    def from_model_on_disk(cls, mod: ModelOnDisk, override_fields: dict[str, Any]) -> Self:
        raise_if_not_file(mod)

        raise_for_override_fields(cls, override_fields)

        if not cls._file_looks_like_embedding(mod):
            raise NotAMatchError("model does not look like a textual inversion embedding file")

        cls._validate_base(mod)

        return cls(**override_fields)


class TI_File_SD1_Config(TI_File_Config_Base, Config_Base):
    base: Literal[BaseModelType.StableDiffusion1] = Field(default=BaseModelType.StableDiffusion1)


class TI_File_SD2_Config(TI_File_Config_Base, Config_Base):
    base: Literal[BaseModelType.StableDiffusion2] = Field(default=BaseModelType.StableDiffusion2)


class TI_File_SDXL_Config(TI_File_Config_Base, Config_Base):
    base: Literal[BaseModelType.StableDiffusionXL] = Field(default=BaseModelType.StableDiffusionXL)


class TI_Folder_Config_Base(TI_Config_Base):
    """Model config for textual inversion embeddings."""

    format: Literal[ModelFormat.EmbeddingFolder] = Field(default=ModelFormat.EmbeddingFolder)

    @classmethod
    def from_model_on_disk(cls, mod: ModelOnDisk, override_fields: dict[str, Any]) -> Self:
        raise_if_not_dir(mod)

        raise_for_override_fields(cls, override_fields)

        for p in mod.weight_files():
            if cls._file_looks_like_embedding(mod, p):
                cls._validate_base(mod, p)
                return cls(**override_fields)

        raise NotAMatchError("model does not look like a textual inversion embedding folder")


class TI_Folder_SD1_Config(TI_Folder_Config_Base, Config_Base):
    base: Literal[BaseModelType.StableDiffusion1] = Field(default=BaseModelType.StableDiffusion1)


class TI_Folder_SD2_Config(TI_Folder_Config_Base, Config_Base):
    base: Literal[BaseModelType.StableDiffusion2] = Field(default=BaseModelType.StableDiffusion2)


class TI_Folder_SDXL_Config(TI_Folder_Config_Base, Config_Base):
    base: Literal[BaseModelType.StableDiffusionXL] = Field(default=BaseModelType.StableDiffusionXL)



import torch
from torch import Tensor
from typing_extensions import Self


class LoraConversionKeySet:
    def __init__(
        self,
        omi_prefix: str,
        diffusers_prefix: str,
        legacy_diffusers_prefix: str | None = None,
        parent: Self | None = None,
        swap_chunks: bool = False,
        filter_is_last: bool | None = None,
        next_omi_prefix: str | None = None,
        next_diffusers_prefix: str | None = None,
    ):
        if parent is not None:
            self.omi_prefix = combine(parent.omi_prefix, omi_prefix)
            self.diffusers_prefix = combine(parent.diffusers_prefix, diffusers_prefix)
        else:
            self.omi_prefix = omi_prefix
            self.diffusers_prefix = diffusers_prefix

        if legacy_diffusers_prefix is None:
            self.legacy_diffusers_prefix = self.diffusers_prefix.replace(".", "_")
        elif parent is not None:
            self.legacy_diffusers_prefix = combine(parent.legacy_diffusers_prefix, legacy_diffusers_prefix).replace(
                ".", "_"
            )
        else:
            self.legacy_diffusers_prefix = legacy_diffusers_prefix

        self.parent = parent
        self.swap_chunks = swap_chunks
        self.filter_is_last = filter_is_last
        self.prefix = parent

        if next_omi_prefix is None and parent is not None:
            self.next_omi_prefix = parent.next_omi_prefix
            self.next_diffusers_prefix = parent.next_diffusers_prefix
            self.next_legacy_diffusers_prefix = parent.next_legacy_diffusers_prefix
        elif next_omi_prefix is not None and parent is not None:
            self.next_omi_prefix = combine(parent.omi_prefix, next_omi_prefix)
            self.next_diffusers_prefix = combine(parent.diffusers_prefix, next_diffusers_prefix)
            self.next_legacy_diffusers_prefix = combine(parent.legacy_diffusers_prefix, next_diffusers_prefix).replace(
                ".", "_"
            )
        elif next_omi_prefix is not None and parent is None:
            self.next_omi_prefix = next_omi_prefix
            self.next_diffusers_prefix = next_diffusers_prefix
            self.next_legacy_diffusers_prefix = next_diffusers_prefix.replace(".", "_")
        else:
            self.next_omi_prefix = None
            self.next_diffusers_prefix = None
            self.next_legacy_diffusers_prefix = None

    def __get_omi(self, in_prefix: str, key: str) -> str:
        return self.omi_prefix + key.removeprefix(in_prefix)

    def __get_diffusers(self, in_prefix: str, key: str) -> str:
        return self.diffusers_prefix + key.removeprefix(in_prefix)

    def __get_legacy_diffusers(self, in_prefix: str, key: str) -> str:
        key = self.legacy_diffusers_prefix + key.removeprefix(in_prefix)

        suffix = key[key.rfind(".") :]
        if suffix not in [".alpha", ".dora_scale"]:  # some keys only have a single . in the suffix
            suffix = key[key.removesuffix(suffix).rfind(".") :]
        key = key.removesuffix(suffix)

        return key.replace(".", "_") + suffix

    def get_key(self, in_prefix: str, key: str, target: str) -> str:
        if target == "omi":
            return self.__get_omi(in_prefix, key)
        elif target == "diffusers":
            return self.__get_diffusers(in_prefix, key)
        elif target == "legacy_diffusers":
            return self.__get_legacy_diffusers(in_prefix, key)
        return key

    def __str__(self) -> str:
        return f"omi: {self.omi_prefix}, diffusers: {self.diffusers_prefix}, legacy: {self.legacy_diffusers_prefix}"


def combine(left: str, right: str) -> str:
    left = left.rstrip(".")
    right = right.lstrip(".")
    if left == "" or left is None:
        return right
    elif right == "" or right is None:
        return left
    else:
        return left + "." + right


def map_prefix_range(
    omi_prefix: str,
    diffusers_prefix: str,
    parent: LoraConversionKeySet,
) -> list[LoraConversionKeySet]:
    # 100 should be a safe upper bound. increase if it's not enough in the future
    return [
        LoraConversionKeySet(
            omi_prefix=f"{omi_prefix}.{i}",
            diffusers_prefix=f"{diffusers_prefix}.{i}",
            parent=parent,
            next_omi_prefix=f"{omi_prefix}.{i + 1}",
            next_diffusers_prefix=f"{diffusers_prefix}.{i + 1}",
        )
        for i in range(100)
    ]


def __convert(
    state_dict: dict[str, Tensor],
    key_sets: list[LoraConversionKeySet],
    source: str,
    target: str,
) -> dict[str, Tensor]:
    out_states = {}

    if source == target:
        return dict(state_dict)

    # TODO: maybe replace with a non O(n^2) algorithm
    for key, tensor in state_dict.items():
        for key_set in key_sets:
            in_prefix = ""

            if source == "omi":
                in_prefix = key_set.omi_prefix
            elif source == "diffusers":
                in_prefix = key_set.diffusers_prefix
            elif source == "legacy_diffusers":
                in_prefix = key_set.legacy_diffusers_prefix

            if not key.startswith(in_prefix):
                continue

            if key_set.filter_is_last is not None:
                next_prefix = None
                if source == "omi":
                    next_prefix = key_set.next_omi_prefix
                elif source == "diffusers":
                    next_prefix = key_set.next_diffusers_prefix
                elif source == "legacy_diffusers":
                    next_prefix = key_set.next_legacy_diffusers_prefix

                is_last = not any(k.startswith(next_prefix) for k in state_dict)
                if key_set.filter_is_last != is_last:
                    continue

            name = key_set.get_key(in_prefix, key, target)

            can_swap_chunks = target == "omi" or source == "omi"
            if key_set.swap_chunks and name.endswith(".lora_up.weight") and can_swap_chunks:
                chunk_0, chunk_1 = tensor.chunk(2, dim=0)
                tensor = torch.cat([chunk_1, chunk_0], dim=0)

            out_states[name] = tensor

            break  # only map the first matching key set

    return out_states


def __detect_source(
    state_dict: dict[str, Tensor],
    key_sets: list[LoraConversionKeySet],
) -> str:
    omi_count = 0
    diffusers_count = 0
    legacy_diffusers_count = 0

    for key in state_dict:
        for key_set in key_sets:
            if key.startswith(key_set.omi_prefix):
                omi_count += 1
            if key.startswith(key_set.diffusers_prefix):
                diffusers_count += 1
            if key.startswith(key_set.legacy_diffusers_prefix):
                legacy_diffusers_count += 1

    if omi_count > diffusers_count and omi_count > legacy_diffusers_count:
        return "omi"
    if diffusers_count > omi_count and diffusers_count > legacy_diffusers_count:
        return "diffusers"
    if legacy_diffusers_count > omi_count and legacy_diffusers_count > diffusers_count:
        return "legacy_diffusers"

    return ""


def convert_to_omi(
    state_dict: dict[str, Tensor],
    key_sets: list[LoraConversionKeySet],
) -> dict[str, Tensor]:
    source = __detect_source(state_dict, key_sets)
    return __convert(state_dict, key_sets, source, "omi")


def convert_to_diffusers(
    state_dict: dict[str, Tensor],
    key_sets: list[LoraConversionKeySet],
) -> dict[str, Tensor]:
    source = __detect_source(state_dict, key_sets)
    return __convert(state_dict, key_sets, source, "diffusers")


def convert_to_legacy_diffusers(
    state_dict: dict[str, Tensor],
    key_sets: list[LoraConversionKeySet],
) -> dict[str, Tensor]:
    source = __detect_source(state_dict, key_sets)
    return __convert(state_dict, key_sets, source, "legacy_diffusers")



import json

from invokeai.app.services.client_state_persistence.client_state_persistence_base import ClientStatePersistenceABC
from invokeai.app.services.invoker import Invoker
from invokeai.app.services.shared.sqlite.sqlite_database import SqliteDatabase


class ClientStatePersistenceSqlite(ClientStatePersistenceABC):
    """
    Base class for client persistence implementations.
    This class defines the interface for persisting client data.
    """

    def __init__(self, db: SqliteDatabase) -> None:
        super().__init__()
        self._db = db
        self._default_row_id = 1

    def start(self, invoker: Invoker) -> None:
        self._invoker = invoker

    def _get(self) -> dict[str, str] | None:
        with self._db.transaction() as cursor:
            cursor.execute(
                f"""
                SELECT data FROM client_state
                WHERE id = {self._default_row_id}
                """
            )
            row = cursor.fetchone()
            if row is None:
                return None
            return json.loads(row[0])

    def set_by_key(self, queue_id: str, key: str, value: str) -> str:
        state = self._get() or {}
        state.update({key: value})

        with self._db.transaction() as cursor:
            cursor.execute(
                f"""
                INSERT INTO client_state (id, data)
                VALUES ({self._default_row_id}, ?)
                ON CONFLICT(id) DO UPDATE
                  SET data = excluded.data;
                """,
                (json.dumps(state),),
            )

        return value

    def get_by_key(self, queue_id: str, key: str) -> str | None:
        state = self._get()
        if state is None:
            return None
        return state.get(key, None)

    def delete(self, queue_id: str) -> None:
        with self._db.transaction() as cursor:
            cursor.execute(
                f"""
                DELETE FROM client_state
                WHERE id = {self._default_row_id}
                """
            )



from invokeai.backend.model_manager.omi.omi import convert_from_omi
from invokeai.backend.model_manager.omi.vendor.model_spec.architecture import (
    flux_dev_1_lora,
    stable_diffusion_xl_1_lora,
)

__all__ = ["flux_dev_1_lora", "stable_diffusion_xl_1_lora", "convert_from_omi"]



from enum import Enum

from pydantic import BaseModel, model_validator
from pydantic.fields import Field


class BoundingBox(BaseModel):
    x_min: int = Field(..., description="The minimum x-coordinate of the bounding box (inclusive).")
    x_max: int = Field(..., description="The maximum x-coordinate of the bounding box (exclusive).")
    y_min: int = Field(..., description="The minimum y-coordinate of the bounding box (inclusive).")
    y_max: int = Field(..., description="The maximum y-coordinate of the bounding box (exclusive).")

    @model_validator(mode="after")
    def check_coords(self):
        if self.x_min > self.x_max:
            raise ValueError(f"x_min ({self.x_min}) is greater than x_max ({self.x_max}).")
        if self.y_min > self.y_max:
            raise ValueError(f"y_min ({self.y_min}) is greater than y_max ({self.y_max}).")
        return self

    def tuple(self) -> tuple[int, int, int, int]:
        """
        Returns the bounding box as a tuple suitable for use with PIL's `Image.crop()` method.
        This method returns a tuple of the form (left, upper, right, lower) == (x_min, y_min, x_max, y_max).
        """
        return (self.x_min, self.y_min, self.x_max, self.y_max)


class SAMPointLabel(Enum):
    negative = -1
    neutral = 0
    positive = 1


class SAMPoint(BaseModel):
    x: int = Field(..., description="The x-coordinate of the point")
    y: int = Field(..., description="The y-coordinate of the point")
    label: SAMPointLabel = Field(..., description="The label of the point")


class SAMInput(BaseModel):
    bounding_box: BoundingBox | None = Field(None, description="The bounding box to use for segmentation")
    points: list[SAMPoint] | None = Field(None, description="The points to use for segmentation")

    @model_validator(mode="after")
    def check_input(self):
        if not self.bounding_box and not self.points:
            raise ValueError("Either bounding_box or points must be provided")
        return self



from typing import Any

import torch
from PIL.Image import Image
from pydantic import field_validator
from transformers import AutoProcessor, LlavaOnevisionForConditionalGeneration, LlavaOnevisionProcessor

from invokeai.app.invocations.baseinvocation import BaseInvocation, Classification, invocation
from invokeai.app.invocations.fields import FieldDescriptions, ImageField, InputField, UIComponent
from invokeai.app.invocations.model import ModelIdentifierField
from invokeai.app.invocations.primitives import StringOutput
from invokeai.app.services.shared.invocation_context import InvocationContext
from invokeai.backend.llava_onevision_pipeline import LlavaOnevisionPipeline
from invokeai.backend.model_manager.taxonomy import ModelType
from invokeai.backend.util.devices import TorchDevice


@invocation(
    "llava_onevision_vllm",
    title="LLaVA OneVision VLLM",
    tags=["vllm"],
    category="vllm",
    version="1.0.0",
    classification=Classification.Beta,
)
class LlavaOnevisionVllmInvocation(BaseInvocation):
    """Run a LLaVA OneVision VLLM model."""

    images: list[ImageField] | ImageField | None = InputField(default=None, max_length=3, description="Input image.")
    prompt: str = InputField(
        default="",
        description="Input text prompt.",
        ui_component=UIComponent.Textarea,
    )
    vllm_model: ModelIdentifierField = InputField(
        title="LLaVA Model Type",
        description=FieldDescriptions.vllm_model,
        ui_model_type=ModelType.LlavaOnevision,
    )

    @field_validator("images", mode="before")
    def listify_images(cls, v: Any) -> list:
        if v is None:
            return v
        if not isinstance(v, list):
            return [v]
        return v

    def _get_images(self, context: InvocationContext) -> list[Image]:
        if self.images is None:
            return []

        image_fields = self.images if isinstance(self.images, list) else [self.images]
        return [context.images.get_pil(image_field.image_name, "RGB") for image_field in image_fields]

    @torch.no_grad()
    def invoke(self, context: InvocationContext) -> StringOutput:
        images = self._get_images(context)
        model_config = context.models.get_config(self.vllm_model)

        with context.models.load(self.vllm_model).model_on_device() as (_, model):
            assert isinstance(model, LlavaOnevisionForConditionalGeneration)

            model_abs_path = context.models.get_absolute_path(model_config)
            processor = AutoProcessor.from_pretrained(model_abs_path, local_files_only=True)
            assert isinstance(processor, LlavaOnevisionProcessor)

            model = LlavaOnevisionPipeline(model, processor)
            output = model.run(
                prompt=self.prompt,
                images=images,
                device=TorchDevice.choose_torch_device(),
                dtype=TorchDevice.choose_torch_dtype(),
            )

        return StringOutput(value=output)



from fastapi import Body, HTTPException, Path, Query
from fastapi.routing import APIRouter

from invokeai.app.api.dependencies import ApiDependencies
from invokeai.backend.util.logging import logging

client_state_router = APIRouter(prefix="/v1/client_state", tags=["client_state"])


@client_state_router.get(
    "/{queue_id}/get_by_key",
    operation_id="get_client_state_by_key",
    response_model=str | None,
)
async def get_client_state_by_key(
    queue_id: str = Path(description="The queue id to perform this operation on"),
    key: str = Query(..., description="Key to get"),
) -> str | None:
    """Gets the client state"""
    try:
        return ApiDependencies.invoker.services.client_state_persistence.get_by_key(queue_id, key)
    except Exception as e:
        logging.error(f"Error getting client state: {e}")
        raise HTTPException(status_code=500, detail="Error setting client state")


@client_state_router.post(
    "/{queue_id}/set_by_key",
    operation_id="set_client_state",
    response_model=str,
)
async def set_client_state(
    queue_id: str = Path(description="The queue id to perform this operation on"),
    key: str = Query(..., description="Key to set"),
    value: str = Body(..., description="Stringified value to set"),
) -> str:
    """Sets the client state"""
    try:
        return ApiDependencies.invoker.services.client_state_persistence.set_by_key(queue_id, key, value)
    except Exception as e:
        logging.error(f"Error setting client state: {e}")
        raise HTTPException(status_code=500, detail="Error setting client state")


@client_state_router.post(
    "/{queue_id}/delete",
    operation_id="delete_client_state",
    responses={204: {"description": "Client state deleted"}},
)
async def delete_client_state(
    queue_id: str = Path(description="The queue id to perform this operation on"),
) -> None:
    """Deletes the client state"""
    try:
        ApiDependencies.invoker.services.client_state_persistence.delete(queue_id)
    except Exception as e:
        logging.error(f"Error deleting client state: {e}")
        raise HTTPException(status_code=500, detail="Error deleting client state")



from typing import (
    Literal,
    Self,
)

from pydantic import Field
from typing_extensions import Any

from invokeai.backend.model_manager.configs.base import Config_Base, Diffusers_Config_Base
from invokeai.backend.model_manager.configs.identification_utils import (
    common_config_paths,
    raise_for_class_name,
    raise_for_override_fields,
    raise_if_not_dir,
)
from invokeai.backend.model_manager.model_on_disk import ModelOnDisk
from invokeai.backend.model_manager.taxonomy import (
    BaseModelType,
    ModelType,
)


class LlavaOnevision_Diffusers_Config(Diffusers_Config_Base, Config_Base):
    """Model config for Llava Onevision models."""

    type: Literal[ModelType.LlavaOnevision] = Field(default=ModelType.LlavaOnevision)
    base: Literal[BaseModelType.Any] = Field(default=BaseModelType.Any)

    @classmethod
    def from_model_on_disk(cls, mod: ModelOnDisk, override_fields: dict[str, Any]) -> Self:
        raise_if_not_dir(mod)

        raise_for_override_fields(cls, override_fields)

        raise_for_class_name(
            common_config_paths(mod.path),
            {
                "LlavaOnevisionForConditionalGeneration",
            },
        )

        return cls(**override_fields)



from typing import Optional

import torch
from PIL import Image

# Import SAM2 components - these should be available in transformers 4.56.0+
from transformers.models.sam2 import Sam2Model
from transformers.models.sam2.processing_sam2 import Sam2Processor

from invokeai.backend.image_util.segment_anything.shared import SAMInput
from invokeai.backend.raw_model import RawModel


class SegmentAnything2Pipeline(RawModel):
    """A wrapper class for the transformers SAM2 model and processor that makes it compatible with the model manager."""

    def __init__(self, sam2_model: Sam2Model, sam2_processor: Sam2Processor):
        """Initialize the SAM2 pipeline.

        Args:
            sam2_model: The SAM2 model
            sam2_processor: The SAM2 processor (can be Sam2Processor or Sam2VideoProcessor)
        """
        self._sam2_model = sam2_model
        self._sam2_processor = sam2_processor

    def to(self, device: Optional[torch.device] = None, dtype: Optional[torch.dtype] = None):
        # HACK: The SAM2 pipeline may not work on MPS devices. We only allow it to be moved to CPU or CUDA.
        if device is not None and device.type not in {"cpu", "cuda"}:
            device = None
        self._sam2_model.to(device=device, dtype=dtype)

    def calc_size(self) -> int:
        # HACK: Fix the circular import issue.
        from invokeai.backend.model_manager.load.model_util import calc_module_size

        return calc_module_size(self._sam2_model)

    def segment(
        self,
        image: Image.Image,
        inputs: list[SAMInput],
    ) -> torch.Tensor:
        """Segment the image using the provided inputs.

        Args:
            image: The image to segment.
            inputs: A list of SAMInput objects containing bounding boxes and/or point lists.

        Returns:
            torch.Tensor: The segmentation masks. dtype: torch.bool. shape: [num_masks, channels, height, width].
        """

        input_boxes: list[list[float]] = []
        input_points: list[list[list[float]]] = []
        input_labels: list[list[int]] = []

        for i in inputs:
            box: list[float] | None = None
            points: list[list[float]] | None = None
            labels: list[int] | None = None

            if i.bounding_box is not None:
                box: list[float] | None = [
                    i.bounding_box.x_min,
                    i.bounding_box.y_min,
                    i.bounding_box.x_max,
                    i.bounding_box.y_max,
                ]

            if i.points is not None:
                points = []
                labels = []
                for point in i.points:
                    points.append([point.x, point.y])
                    labels.append(point.label.value)

            if box is not None:
                input_boxes.append(box)
            if points is not None:
                input_points.append(points)
            if labels is not None:
                input_labels.append(labels)

        batched_input_boxes = [input_boxes] if input_boxes else None
        batched_input_points = [input_points] if input_points else None
        batched_input_labels = [input_labels] if input_labels else None

        processed_inputs = self._sam2_processor(
            images=image,
            input_boxes=batched_input_boxes,
            input_points=batched_input_points,
            input_labels=batched_input_labels,
            return_tensors="pt",
        ).to(self._sam2_model.device)

        # Generate masks using the SAM2 model
        outputs = self._sam2_model(**processed_inputs)

        # Post-process the masks to get the final segmentation
        masks = self._sam2_processor.post_process_masks(
            masks=outputs.pred_masks,
            original_sizes=processed_inputs.original_sizes,
            reshaped_input_sizes=processed_inputs.reshaped_input_sizes,
        )

        # There should be only one batch.
        assert len(masks) == 1
        return masks[0]



from typing import Any, Literal, Self

from pydantic import Field

from invokeai.backend.model_manager.configs.base import Config_Base
from invokeai.backend.model_manager.configs.identification_utils import (
    NotAMatchError,
    raise_for_class_name,
    raise_for_override_fields,
    raise_if_not_dir,
    state_dict_has_any_keys_ending_with,
)
from invokeai.backend.model_manager.model_on_disk import ModelOnDisk
from invokeai.backend.model_manager.taxonomy import BaseModelType, ModelFormat, ModelType


class T5Encoder_T5Encoder_Config(Config_Base):
    """Configuration for T5 Encoder models in a bespoke, diffusers-like format. The model weights are expected to be in
    a folder called text_encoder_2 inside the model directory, with a config file named model.safetensors.index.json."""

    base: Literal[BaseModelType.Any] = Field(default=BaseModelType.Any)
    type: Literal[ModelType.T5Encoder] = Field(default=ModelType.T5Encoder)
    format: Literal[ModelFormat.T5Encoder] = Field(default=ModelFormat.T5Encoder)

    @classmethod
    def from_model_on_disk(cls, mod: ModelOnDisk, override_fields: dict[str, Any]) -> Self:
        raise_if_not_dir(mod)

        raise_for_override_fields(cls, override_fields)

        expected_config_path = mod.path / "text_encoder_2" / "config.json"
        expected_class_name = "T5EncoderModel"
        raise_for_class_name(expected_config_path, expected_class_name)

        cls.raise_if_doesnt_have_unquantized_config_file(mod)

        return cls(**override_fields)

    @classmethod
    def raise_if_doesnt_have_unquantized_config_file(cls, mod: ModelOnDisk) -> None:
        has_unquantized_config = (mod.path / "text_encoder_2" / "model.safetensors.index.json").exists()

        if not has_unquantized_config:
            raise NotAMatchError("missing text_encoder_2/model.safetensors.index.json")


class T5Encoder_BnBLLMint8_Config(Config_Base):
    """Configuration for T5 Encoder models quantized by bitsandbytes' LLM.int8."""

    base: Literal[BaseModelType.Any] = Field(default=BaseModelType.Any)
    type: Literal[ModelType.T5Encoder] = Field(default=ModelType.T5Encoder)
    format: Literal[ModelFormat.BnbQuantizedLlmInt8b] = Field(default=ModelFormat.BnbQuantizedLlmInt8b)

    @classmethod
    def from_model_on_disk(cls, mod: ModelOnDisk, override_fields: dict[str, Any]) -> Self:
        raise_if_not_dir(mod)

        raise_for_override_fields(cls, override_fields)

        expected_config_path = mod.path / "text_encoder_2" / "config.json"
        expected_class_name = "T5EncoderModel"
        raise_for_class_name(expected_config_path, expected_class_name)

        cls.raise_if_filename_doesnt_look_like_bnb_quantized(mod)

        cls.raise_if_state_dict_doesnt_look_like_bnb_quantized(mod)

        return cls(**override_fields)

    @classmethod
    def raise_if_filename_doesnt_look_like_bnb_quantized(cls, mod: ModelOnDisk) -> None:
        filename_looks_like_bnb = any(x for x in mod.weight_files() if "llm_int8" in x.as_posix())
        if not filename_looks_like_bnb:
            raise NotAMatchError("filename does not look like bnb quantized llm_int8")

    @classmethod
    def raise_if_state_dict_doesnt_look_like_bnb_quantized(cls, mod: ModelOnDisk) -> None:
        has_scb_key_suffix = state_dict_has_any_keys_ending_with(mod.load_state_dict(), "SCB")
        if not has_scb_key_suffix:
            raise NotAMatchError("state dict does not look like bnb quantized llm_int8")



from typing import (
    Literal,
    Self,
)

from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import Any

from invokeai.backend.flux.controlnet.state_dict_utils import (
    is_state_dict_instantx_controlnet,
    is_state_dict_xlabs_controlnet,
)
from invokeai.backend.model_manager.configs.base import Checkpoint_Config_Base, Config_Base, Diffusers_Config_Base
from invokeai.backend.model_manager.configs.identification_utils import (
    NotAMatchError,
    common_config_paths,
    get_config_dict_or_raise,
    raise_for_class_name,
    raise_for_override_fields,
    raise_if_not_dir,
    raise_if_not_file,
    state_dict_has_any_keys_starting_with,
)
from invokeai.backend.model_manager.model_on_disk import ModelOnDisk
from invokeai.backend.model_manager.taxonomy import (
    BaseModelType,
    ModelFormat,
    ModelType,
)

MODEL_NAME_TO_PREPROCESSOR = {
    "canny": "canny_image_processor",
    "mlsd": "mlsd_image_processor",
    "depth": "depth_anything_image_processor",
    "bae": "normalbae_image_processor",
    "normal": "normalbae_image_processor",
    "sketch": "pidi_image_processor",
    "scribble": "lineart_image_processor",
    "lineart anime": "lineart_anime_image_processor",
    "lineart_anime": "lineart_anime_image_processor",
    "lineart": "lineart_image_processor",
    "soft": "hed_image_processor",
    "softedge": "hed_image_processor",
    "hed": "hed_image_processor",
    "shuffle": "content_shuffle_image_processor",
    "pose": "dw_openpose_image_processor",
    "mediapipe": "mediapipe_face_processor",
    "pidi": "pidi_image_processor",
    "zoe": "zoe_depth_image_processor",
    "color": "color_map_image_processor",
}


class ControlAdapterDefaultSettings(BaseModel):
    # This could be narrowed to controlnet processor nodes, but they change. Leaving this a string is safer.
    preprocessor: str | None
    model_config = ConfigDict(extra="forbid")

    @classmethod
    def from_model_name(cls, model_name: str) -> Self:
        for k, v in MODEL_NAME_TO_PREPROCESSOR.items():
            model_name_lower = model_name.lower()
            if k in model_name_lower:
                return cls(preprocessor=v)
        return cls(preprocessor=None)


class ControlNet_Diffusers_Config_Base(Diffusers_Config_Base):
    """Model config for ControlNet models (diffusers version)."""

    type: Literal[ModelType.ControlNet] = Field(default=ModelType.ControlNet)
    format: Literal[ModelFormat.Diffusers] = Field(default=ModelFormat.Diffusers)
    default_settings: ControlAdapterDefaultSettings | None = Field(None)

    @classmethod
    def from_model_on_disk(cls, mod: ModelOnDisk, override_fields: dict[str, Any]) -> Self:
        raise_if_not_dir(mod)

        raise_for_override_fields(cls, override_fields)

        raise_for_class_name(
            common_config_paths(mod.path),
            {
                "ControlNetModel",
                "FluxControlNetModel",
            },
        )

        cls._validate_base(mod)

        return cls(**override_fields)

    @classmethod
    def _validate_base(cls, mod: ModelOnDisk) -> None:
        """Raise `NotAMatch` if the model base does not match this config class."""
        expected_base = cls.model_fields["base"].default
        recognized_base = cls._get_base_or_raise(mod)
        if expected_base is not recognized_base:
            raise NotAMatchError(f"base is {recognized_base}, not {expected_base}")

    @classmethod
    def _get_base_or_raise(cls, mod: ModelOnDisk) -> BaseModelType:
        config_dict = get_config_dict_or_raise(common_config_paths(mod.path))

        if config_dict.get("_class_name") == "FluxControlNetModel":
            return BaseModelType.Flux

        dimension = config_dict.get("cross_attention_dim")

        match dimension:
            case 768:
                return BaseModelType.StableDiffusion1
            case 1024:
                # No obvious way to distinguish between sd2-base and sd2-768, but we don't really differentiate them
                # anyway.
                return BaseModelType.StableDiffusion2
            case 2048:
                return BaseModelType.StableDiffusionXL
            case _:
                raise NotAMatchError(f"unrecognized cross_attention_dim {dimension}")


class ControlNet_Diffusers_SD1_Config(ControlNet_Diffusers_Config_Base, Config_Base):
    base: Literal[BaseModelType.StableDiffusion1] = Field(default=BaseModelType.StableDiffusion1)


class ControlNet_Diffusers_SD2_Config(ControlNet_Diffusers_Config_Base, Config_Base):
    base: Literal[BaseModelType.StableDiffusion2] = Field(default=BaseModelType.StableDiffusion2)


class ControlNet_Diffusers_SDXL_Config(ControlNet_Diffusers_Config_Base, Config_Base):
    base: Literal[BaseModelType.StableDiffusionXL] = Field(default=BaseModelType.StableDiffusionXL)


class ControlNet_Diffusers_FLUX_Config(ControlNet_Diffusers_Config_Base, Config_Base):
    base: Literal[BaseModelType.Flux] = Field(default=BaseModelType.Flux)


class ControlNet_Checkpoint_Config_Base(Checkpoint_Config_Base):
    """Model config for ControlNet models (diffusers version)."""

    type: Literal[ModelType.ControlNet] = Field(default=ModelType.ControlNet)
    format: Literal[ModelFormat.Checkpoint] = Field(default=ModelFormat.Checkpoint)
    default_settings: ControlAdapterDefaultSettings | None = Field(None)

    @classmethod
    def from_model_on_disk(cls, mod: ModelOnDisk, override_fields: dict[str, Any]) -> Self:
        raise_if_not_file(mod)

        raise_for_override_fields(cls, override_fields)

        cls._validate_looks_like_controlnet(mod)

        cls._validate_base(mod)

        return cls(**override_fields)

    @classmethod
    def _validate_base(cls, mod: ModelOnDisk) -> None:
        """Raise `NotAMatch` if the model base does not match this config class."""
        expected_base = cls.model_fields["base"].default
        recognized_base = cls._get_base_or_raise(mod)
        if expected_base is not recognized_base:
            raise NotAMatchError(f"base is {recognized_base}, not {expected_base}")

    @classmethod
    def _validate_looks_like_controlnet(cls, mod: ModelOnDisk) -> None:
        if not state_dict_has_any_keys_starting_with(
            mod.load_state_dict(),
            {
                "controlnet",
                "control_model",
                "input_blocks",
                # XLabs FLUX ControlNet models have keys starting with "controlnet_blocks."
                # For example: https://huggingface.co/XLabs-AI/flux-controlnet-collections/blob/86ab1e915a389d5857135c00e0d350e9e38a9048/flux-canny-controlnet_v2.safetensors
                # TODO(ryand): This is very fragile. XLabs FLUX ControlNet models also contain keys starting with
                # "double_blocks.", which we check for above. But, I'm afraid to modify this logic because it is so
                # delicate.
                "controlnet_blocks",
            },
        ):
            raise NotAMatchError("state dict does not look like a ControlNet checkpoint")

    @classmethod
    def _get_base_or_raise(cls, mod: ModelOnDisk) -> BaseModelType:
        state_dict = mod.load_state_dict()

        if is_state_dict_xlabs_controlnet(state_dict) or is_state_dict_instantx_controlnet(state_dict):
            # TODO(ryand): Should I distinguish between XLabs, InstantX and other ControlNet models by implementing
            # get_format()?
            return BaseModelType.Flux

        for key in (
            "control_model.input_blocks.2.1.transformer_blocks.0.attn2.to_k.weight",
            "controlnet_mid_block.bias",
            "input_blocks.2.1.transformer_blocks.0.attn2.to_k.weight",
            "down_blocks.1.attentions.0.transformer_blocks.0.attn2.to_k.weight",
        ):
            if key not in state_dict:
                continue
            width = state_dict[key].shape[-1]
            match width:
                case 768:
                    return BaseModelType.StableDiffusion1
                case 1024:
                    return BaseModelType.StableDiffusion2
                case 2048:
                    return BaseModelType.StableDiffusionXL
                case 1280:
                    return BaseModelType.StableDiffusionXL
                case _:
                    pass

        raise NotAMatchError("unable to determine base type from state dict")


class ControlNet_Checkpoint_SD1_Config(ControlNet_Checkpoint_Config_Base, Config_Base):
    base: Literal[BaseModelType.StableDiffusion1] = Field(default=BaseModelType.StableDiffusion1)


class ControlNet_Checkpoint_SD2_Config(ControlNet_Checkpoint_Config_Base, Config_Base):
    base: Literal[BaseModelType.StableDiffusion2] = Field(default=BaseModelType.StableDiffusion2)


class ControlNet_Checkpoint_SDXL_Config(ControlNet_Checkpoint_Config_Base, Config_Base):
    base: Literal[BaseModelType.StableDiffusionXL] = Field(default=BaseModelType.StableDiffusionXL)


class ControlNet_Checkpoint_FLUX_Config(ControlNet_Checkpoint_Config_Base, Config_Base):
    base: Literal[BaseModelType.Flux] = Field(default=BaseModelType.Flux)



import sqlite3

from invokeai.app.services.shared.sqlite_migrator.sqlite_migrator_common import Migration


class Migration21Callback:
    def __call__(self, cursor: sqlite3.Cursor) -> None:
        cursor.execute(
            """
            CREATE TABLE client_state (
              id          INTEGER PRIMARY KEY CHECK(id = 1),
              data        TEXT    NOT NULL, -- Frontend will handle the shape of this data
              updated_at  DATETIME NOT NULL DEFAULT (CURRENT_TIMESTAMP)
            );
            """
        )
        cursor.execute(
            """
            CREATE TRIGGER tg_client_state_updated_at
            AFTER UPDATE ON client_state
            FOR EACH ROW
            BEGIN
              UPDATE client_state
                SET updated_at = CURRENT_TIMESTAMP
              WHERE id = OLD.id;
            END;
            """
        )


def build_migration_21() -> Migration:
    """Builds the migration object for migrating from version 20 to version 21. This includes:
    - Creating the `client_state` table.
    - Adding a trigger to update the `updated_at` field on updates.
    """
    return Migration(
        from_version=20,
        to_version=21,
        callback=Migration21Callback(),
    )



import sqlite3
from datetime import datetime
from typing import Optional, Union, cast

from invokeai.app.invocations.fields import MetadataField, MetadataFieldValidator
from invokeai.app.services.image_records.image_records_base import ImageRecordStorageBase
from invokeai.app.services.image_records.image_records_common import (
    IMAGE_DTO_COLS,
    ImageCategory,
    ImageNamesResult,
    ImageRecord,
    ImageRecordChanges,
    ImageRecordDeleteException,
    ImageRecordNotFoundException,
    ImageRecordSaveException,
    ResourceOrigin,
    deserialize_image_record,
)
from invokeai.app.services.shared.pagination import OffsetPaginatedResults
from invokeai.app.services.shared.sqlite.sqlite_common import SQLiteDirection
from invokeai.app.services.shared.sqlite.sqlite_database import SqliteDatabase


class SqliteImageRecordStorage(ImageRecordStorageBase):
    def __init__(self, db: SqliteDatabase) -> None:
        super().__init__()
        self._db = db

    def get(self, image_name: str) -> ImageRecord:
        with self._db.transaction() as cursor:
            try:
                cursor.execute(
                    f"""--sql
                    SELECT {IMAGE_DTO_COLS} FROM images
                    WHERE image_name = ?;
                    """,
                    (image_name,),
                )

                result = cast(Optional[sqlite3.Row], cursor.fetchone())
            except sqlite3.Error as e:
                raise ImageRecordNotFoundException from e

        if not result:
            raise ImageRecordNotFoundException

        return deserialize_image_record(dict(result))

    def get_metadata(self, image_name: str) -> Optional[MetadataField]:
        with self._db.transaction() as cursor:
            try:
                cursor.execute(
                    """--sql
                    SELECT metadata FROM images
                    WHERE image_name = ?;
                    """,
                    (image_name,),
                )

                result = cast(Optional[sqlite3.Row], cursor.fetchone())

            except sqlite3.Error as e:
                raise ImageRecordNotFoundException from e

            if not result:
                raise ImageRecordNotFoundException

            as_dict = dict(result)
            metadata_raw = cast(Optional[str], as_dict.get("metadata", None))
            return MetadataFieldValidator.validate_json(metadata_raw) if metadata_raw is not None else None

    def update(
        self,
        image_name: str,
        changes: ImageRecordChanges,
    ) -> None:
        with self._db.transaction() as cursor:
            try:
                # Change the category of the image
                if changes.image_category is not None:
                    cursor.execute(
                        """--sql
                        UPDATE images
                        SET image_category = ?
                        WHERE image_name = ?;
                        """,
                        (changes.image_category, image_name),
                    )

                # Change the session associated with the image
                if changes.session_id is not None:
                    cursor.execute(
                        """--sql
                        UPDATE images
                        SET session_id = ?
                        WHERE image_name = ?;
                        """,
                        (changes.session_id, image_name),
                    )

                # Change the image's `is_intermediate`` flag
                if changes.is_intermediate is not None:
                    cursor.execute(
                        """--sql
                        UPDATE images
                        SET is_intermediate = ?
                        WHERE image_name = ?;
                        """,
                        (changes.is_intermediate, image_name),
                    )

                # Change the image's `starred`` state
                if changes.starred is not None:
                    cursor.execute(
                        """--sql
                        UPDATE images
                        SET starred = ?
                        WHERE image_name = ?;
                        """,
                        (changes.starred, image_name),
                    )

            except sqlite3.Error as e:
                raise ImageRecordSaveException from e

    def get_many(
        self,
        offset: int = 0,
        limit: int = 10,
        starred_first: bool = True,
        order_dir: SQLiteDirection = SQLiteDirection.Descending,
        image_origin: Optional[ResourceOrigin] = None,
        categories: Optional[list[ImageCategory]] = None,
        is_intermediate: Optional[bool] = None,
        board_id: Optional[str] = None,
        search_term: Optional[str] = None,
    ) -> OffsetPaginatedResults[ImageRecord]:
        with self._db.transaction() as cursor:
            # Manually build two queries - one for the count, one for the records
            count_query = """--sql
            SELECT COUNT(*)
            FROM images
            LEFT JOIN board_images ON board_images.image_name = images.image_name
            WHERE 1=1
            """

            images_query = f"""--sql
            SELECT {IMAGE_DTO_COLS}
            FROM images
            LEFT JOIN board_images ON board_images.image_name = images.image_name
            WHERE 1=1
            """

            query_conditions = ""
            query_params: list[Union[int, str, bool]] = []

            if image_origin is not None:
                query_conditions += """--sql
                AND images.image_origin = ?
                """
                query_params.append(image_origin.value)

            if categories is not None:
                # Convert the enum values to unique list of strings
                category_strings = [c.value for c in set(categories)]
                # Create the correct length of placeholders
                placeholders = ",".join("?" * len(category_strings))

                query_conditions += f"""--sql
                AND images.image_category IN ( {placeholders} )
                """

                # Unpack the included categories into the query params
                for c in category_strings:
                    query_params.append(c)

            if is_intermediate is not None:
                query_conditions += """--sql
                AND images.is_intermediate = ?
                """

                query_params.append(is_intermediate)

            # board_id of "none" is reserved for images without a board
            if board_id == "none":
                query_conditions += """--sql
                AND board_images.board_id IS NULL
                """
            elif board_id is not None:
                query_conditions += """--sql
                AND board_images.board_id = ?
                """
                query_params.append(board_id)

            # Search term condition
            if search_term:
                query_conditions += """--sql
                AND (
                    images.metadata LIKE ?
                    OR images.created_at LIKE ?
                )
                """
                query_params.append(f"%{search_term.lower()}%")
                query_params.append(f"%{search_term.lower()}%")

            if starred_first:
                query_pagination = f"""--sql
                ORDER BY images.starred DESC, images.created_at {order_dir.value} LIMIT ? OFFSET ?
                """
            else:
                query_pagination = f"""--sql
                ORDER BY images.created_at {order_dir.value} LIMIT ? OFFSET ?
                """

            # Final images query with pagination
            images_query += query_conditions + query_pagination + ";"
            # Add all the parameters
            images_params = query_params.copy()
            # Add the pagination parameters
            images_params.extend([limit, offset])

            # Build the list of images, deserializing each row
            cursor.execute(images_query, images_params)
            result = cast(list[sqlite3.Row], cursor.fetchall())

            images = [deserialize_image_record(dict(r)) for r in result]

            # Set up and execute the count query, without pagination
            count_query += query_conditions + ";"
            count_params = query_params.copy()
            cursor.execute(count_query, count_params)
            count = cast(int, cursor.fetchone()[0])

        return OffsetPaginatedResults(items=images, offset=offset, limit=limit, total=count)

    def delete(self, image_name: str) -> None:
        with self._db.transaction() as cursor:
            try:
                cursor.execute(
                    """--sql
                    DELETE FROM images
                    WHERE image_name = ?;
                    """,
                    (image_name,),
                )
            except sqlite3.Error as e:
                raise ImageRecordDeleteException from e

    def delete_many(self, image_names: list[str]) -> None:
        with self._db.transaction() as cursor:
            try:
                placeholders = ",".join("?" for _ in image_names)

                # Construct the SQLite query with the placeholders
                query = f"DELETE FROM images WHERE image_name IN ({placeholders})"

                # Execute the query with the list of IDs as parameters
                cursor.execute(query, image_names)

            except sqlite3.Error as e:
                raise ImageRecordDeleteException from e

    def get_intermediates_count(self) -> int:
        with self._db.transaction() as cursor:
            cursor.execute(
                """--sql
                SELECT COUNT(*) FROM images
                WHERE is_intermediate = TRUE;
                """
            )
            count = cast(int, cursor.fetchone()[0])
        return count

    def delete_intermediates(self) -> list[str]:
        with self._db.transaction() as cursor:
            try:
                cursor.execute(
                    """--sql
                    SELECT image_name FROM images
                    WHERE is_intermediate = TRUE;
                    """
                )
                result = cast(list[sqlite3.Row], cursor.fetchall())
                image_names = [r[0] for r in result]
                cursor.execute(
                    """--sql
                    DELETE FROM images
                    WHERE is_intermediate = TRUE;
                    """
                )
            except sqlite3.Error as e:
                raise ImageRecordDeleteException from e
        return image_names

    def save(
        self,
        image_name: str,
        image_origin: ResourceOrigin,
        image_category: ImageCategory,
        width: int,
        height: int,
        has_workflow: bool,
        is_intermediate: Optional[bool] = False,
        starred: Optional[bool] = False,
        session_id: Optional[str] = None,
        node_id: Optional[str] = None,
        metadata: Optional[str] = None,
    ) -> datetime:
        with self._db.transaction() as cursor:
            try:
                cursor.execute(
                    """--sql
                    INSERT OR IGNORE INTO images (
                        image_name,
                        image_origin,
                        image_category,
                        width,
                        height,
                        node_id,
                        session_id,
                        metadata,
                        is_intermediate,
                        starred,
                        has_workflow
                        )
                    VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?);
                    """,
                    (
                        image_name,
                        image_origin.value,
                        image_category.value,
                        width,
                        height,
                        node_id,
                        session_id,
                        metadata,
                        is_intermediate,
                        starred,
                        has_workflow,
                    ),
                )

                cursor.execute(
                    """--sql
                    SELECT created_at
                    FROM images
                    WHERE image_name = ?;
                    """,
                    (image_name,),
                )

                created_at = datetime.fromisoformat(cursor.fetchone()[0])

            except sqlite3.Error as e:
                raise ImageRecordSaveException from e
        return created_at

    def get_most_recent_image_for_board(self, board_id: str) -> Optional[ImageRecord]:
        with self._db.transaction() as cursor:
            cursor.execute(
                """--sql
                SELECT images.*
                FROM images
                JOIN board_images ON images.image_name = board_images.image_name
                WHERE board_images.board_id = ?
                AND images.is_intermediate = FALSE
                ORDER BY images.starred DESC, images.created_at DESC
                LIMIT 1;
                """,
                (board_id,),
            )

            result = cast(Optional[sqlite3.Row], cursor.fetchone())

        if result is None:
            return None

        return deserialize_image_record(dict(result))

    def get_image_names(
        self,
        starred_first: bool = True,
        order_dir: SQLiteDirection = SQLiteDirection.Descending,
        image_origin: Optional[ResourceOrigin] = None,
        categories: Optional[list[ImageCategory]] = None,
        is_intermediate: Optional[bool] = None,
        board_id: Optional[str] = None,
        search_term: Optional[str] = None,
    ) -> ImageNamesResult:
        with self._db.transaction() as cursor:
            # Build query conditions (reused for both starred count and image names queries)
            query_conditions = ""
            query_params: list[Union[int, str, bool]] = []

            if image_origin is not None:
                query_conditions += """--sql
                AND images.image_origin = ?
                """
                query_params.append(image_origin.value)

            if categories is not None:
                category_strings = [c.value for c in set(categories)]
                placeholders = ",".join("?" * len(category_strings))
                query_conditions += f"""--sql
                AND images.image_category IN ( {placeholders} )
                """
                for c in category_strings:
                    query_params.append(c)

            if is_intermediate is not None:
                query_conditions += """--sql
                AND images.is_intermediate = ?
                """
                query_params.append(is_intermediate)

            if board_id == "none":
                query_conditions += """--sql
                AND board_images.board_id IS NULL
                """
            elif board_id is not None:
                query_conditions += """--sql
                AND board_images.board_id = ?
                """
                query_params.append(board_id)

            if search_term:
                query_conditions += """--sql
                AND (
                    images.metadata LIKE ?
                    OR images.created_at LIKE ?
                )
                """
                query_params.append(f"%{search_term.lower()}%")
                query_params.append(f"%{search_term.lower()}%")

            # Get starred count if starred_first is enabled
            starred_count = 0
            if starred_first:
                starred_count_query = f"""--sql
                SELECT COUNT(*)
                FROM images
                LEFT JOIN board_images ON board_images.image_name = images.image_name
                WHERE images.starred = TRUE AND (1=1{query_conditions})
                """
                cursor.execute(starred_count_query, query_params)
                starred_count = cast(int, cursor.fetchone()[0])

            # Get all image names with proper ordering
            if starred_first:
                names_query = f"""--sql
                SELECT images.image_name
                FROM images
                LEFT JOIN board_images ON board_images.image_name = images.image_name
                WHERE 1=1{query_conditions}
                ORDER BY images.starred DESC, images.created_at {order_dir.value}
                """
            else:
                names_query = f"""--sql
                SELECT images.image_name
                FROM images
                LEFT JOIN board_images ON board_images.image_name = images.image_name
                WHERE 1=1{query_conditions}
                ORDER BY images.created_at {order_dir.value}
                """

            cursor.execute(names_query, query_params)
            result = cast(list[sqlite3.Row], cursor.fetchall())
        image_names = [row[0] for row in result]

        return ImageNamesResult(image_names=image_names, starred_count=starred_count, total_count=len(image_names))



import logging
from dataclasses import dataclass
from pathlib import Path
from typing import (
    Union,
)

from pydantic import Discriminator, TypeAdapter, ValidationError
from typing_extensions import Annotated, Any

from invokeai.app.services.config.config_default import get_config
from invokeai.app.util.misc import uuid_string
from invokeai.backend.model_hash.model_hash import HASHING_ALGORITHMS
from invokeai.backend.model_manager.configs.base import Config_Base
from invokeai.backend.model_manager.configs.clip_embed import CLIPEmbed_Diffusers_G_Config, CLIPEmbed_Diffusers_L_Config
from invokeai.backend.model_manager.configs.clip_vision import CLIPVision_Diffusers_Config
from invokeai.backend.model_manager.configs.controlnet import (
    ControlAdapterDefaultSettings,
    ControlNet_Checkpoint_FLUX_Config,
    ControlNet_Checkpoint_SD1_Config,
    ControlNet_Checkpoint_SD2_Config,
    ControlNet_Checkpoint_SDXL_Config,
    ControlNet_Diffusers_FLUX_Config,
    ControlNet_Diffusers_SD1_Config,
    ControlNet_Diffusers_SD2_Config,
    ControlNet_Diffusers_SDXL_Config,
)
from invokeai.backend.model_manager.configs.flux_redux import FLUXRedux_Checkpoint_Config
from invokeai.backend.model_manager.configs.identification_utils import NotAMatchError
from invokeai.backend.model_manager.configs.ip_adapter import (
    IPAdapter_Checkpoint_FLUX_Config,
    IPAdapter_Checkpoint_SD1_Config,
    IPAdapter_Checkpoint_SD2_Config,
    IPAdapter_Checkpoint_SDXL_Config,
    IPAdapter_InvokeAI_SD1_Config,
    IPAdapter_InvokeAI_SD2_Config,
    IPAdapter_InvokeAI_SDXL_Config,
)
from invokeai.backend.model_manager.configs.llava_onevision import LlavaOnevision_Diffusers_Config
from invokeai.backend.model_manager.configs.lora import (
    ControlLoRA_LyCORIS_FLUX_Config,
    LoRA_Diffusers_FLUX_Config,
    LoRA_Diffusers_SD1_Config,
    LoRA_Diffusers_SD2_Config,
    LoRA_Diffusers_SDXL_Config,
    LoRA_LyCORIS_FLUX_Config,
    LoRA_LyCORIS_SD1_Config,
    LoRA_LyCORIS_SD2_Config,
    LoRA_LyCORIS_SDXL_Config,
    LoRA_OMI_FLUX_Config,
    LoRA_OMI_SDXL_Config,
    LoraModelDefaultSettings,
)
from invokeai.backend.model_manager.configs.main import (
    Main_BnBNF4_FLUX_Config,
    Main_Checkpoint_FLUX_Config,
    Main_Checkpoint_SD1_Config,
    Main_Checkpoint_SD2_Config,
    Main_Checkpoint_SDXL_Config,
    Main_Checkpoint_SDXLRefiner_Config,
    Main_Diffusers_CogView4_Config,
    Main_Diffusers_SD1_Config,
    Main_Diffusers_SD2_Config,
    Main_Diffusers_SD3_Config,
    Main_Diffusers_SDXL_Config,
    Main_Diffusers_SDXLRefiner_Config,
    Main_ExternalAPI_ChatGPT4o_Config,
    Main_ExternalAPI_FluxKontext_Config,
    Main_ExternalAPI_Gemini2_5_Config,
    Main_ExternalAPI_Imagen3_Config,
    Main_ExternalAPI_Imagen4_Config,
    Main_GGUF_FLUX_Config,
    MainModelDefaultSettings,
    Video_ExternalAPI_Runway_Config,
    Video_ExternalAPI_Veo3_Config,
)
from invokeai.backend.model_manager.configs.siglip import SigLIP_Diffusers_Config
from invokeai.backend.model_manager.configs.spandrel import Spandrel_Checkpoint_Config
from invokeai.backend.model_manager.configs.t2i_adapter import (
    T2IAdapter_Diffusers_SD1_Config,
    T2IAdapter_Diffusers_SDXL_Config,
)
from invokeai.backend.model_manager.configs.t5_encoder import T5Encoder_BnBLLMint8_Config, T5Encoder_T5Encoder_Config
from invokeai.backend.model_manager.configs.textual_inversion import (
    TI_File_SD1_Config,
    TI_File_SD2_Config,
    TI_File_SDXL_Config,
    TI_Folder_SD1_Config,
    TI_Folder_SD2_Config,
    TI_Folder_SDXL_Config,
)
from invokeai.backend.model_manager.configs.unknown import Unknown_Config
from invokeai.backend.model_manager.configs.vae import (
    VAE_Checkpoint_FLUX_Config,
    VAE_Checkpoint_SD1_Config,
    VAE_Checkpoint_SD2_Config,
    VAE_Checkpoint_SDXL_Config,
    VAE_Diffusers_SD1_Config,
    VAE_Diffusers_SDXL_Config,
)
from invokeai.backend.model_manager.model_on_disk import ModelOnDisk
from invokeai.backend.model_manager.taxonomy import (
    BaseModelType,
    ModelFormat,
    ModelSourceType,
    ModelType,
    variant_type_adapter,
)

logger = logging.getLogger(__name__)
app_config = get_config()

# Known model file extensions for sanity checking
_MODEL_EXTENSIONS = {
    ".safetensors",
    ".ckpt",
    ".pt",
    ".pth",
    ".bin",
    ".gguf",
    ".onnx",
}

# Known config file names for diffusers/transformers models
_CONFIG_FILES = {
    "model_index.json",
    "config.json",
}

# Maximum number of files in a directory to be considered a model
_MAX_FILES_IN_MODEL_DIR = 50

# Maximum depth to search for model files in directories
_MAX_SEARCH_DEPTH = 2


# The types are listed explicitly because IDEs/LSPs can't identify the correct types
# when AnyModelConfig is constructed dynamically using ModelConfigBase.all_config_classes
AnyModelConfig = Annotated[
    Union[
        # Main (Pipeline) - diffusers format
        Annotated[Main_Diffusers_SD1_Config, Main_Diffusers_SD1_Config.get_tag()],
        Annotated[Main_Diffusers_SD2_Config, Main_Diffusers_SD2_Config.get_tag()],
        Annotated[Main_Diffusers_SDXL_Config, Main_Diffusers_SDXL_Config.get_tag()],
        Annotated[Main_Diffusers_SDXLRefiner_Config, Main_Diffusers_SDXLRefiner_Config.get_tag()],
        Annotated[Main_Diffusers_SD3_Config, Main_Diffusers_SD3_Config.get_tag()],
        Annotated[Main_Diffusers_CogView4_Config, Main_Diffusers_CogView4_Config.get_tag()],
        # Main (Pipeline) - checkpoint format
        Annotated[Main_Checkpoint_SD1_Config, Main_Checkpoint_SD1_Config.get_tag()],
        Annotated[Main_Checkpoint_SD2_Config, Main_Checkpoint_SD2_Config.get_tag()],
        Annotated[Main_Checkpoint_SDXL_Config, Main_Checkpoint_SDXL_Config.get_tag()],
        Annotated[Main_Checkpoint_SDXLRefiner_Config, Main_Checkpoint_SDXLRefiner_Config.get_tag()],
        Annotated[Main_Checkpoint_FLUX_Config, Main_Checkpoint_FLUX_Config.get_tag()],
        # Main (Pipeline) - quantized formats
        Annotated[Main_BnBNF4_FLUX_Config, Main_BnBNF4_FLUX_Config.get_tag()],
        Annotated[Main_GGUF_FLUX_Config, Main_GGUF_FLUX_Config.get_tag()],
        # VAE - checkpoint format
        Annotated[VAE_Checkpoint_SD1_Config, VAE_Checkpoint_SD1_Config.get_tag()],
        Annotated[VAE_Checkpoint_SD2_Config, VAE_Checkpoint_SD2_Config.get_tag()],
        Annotated[VAE_Checkpoint_SDXL_Config, VAE_Checkpoint_SDXL_Config.get_tag()],
        Annotated[VAE_Checkpoint_FLUX_Config, VAE_Checkpoint_FLUX_Config.get_tag()],
        # VAE - diffusers format
        Annotated[VAE_Diffusers_SD1_Config, VAE_Diffusers_SD1_Config.get_tag()],
        Annotated[VAE_Diffusers_SDXL_Config, VAE_Diffusers_SDXL_Config.get_tag()],
        # ControlNet - checkpoint format
        Annotated[ControlNet_Checkpoint_SD1_Config, ControlNet_Checkpoint_SD1_Config.get_tag()],
        Annotated[ControlNet_Checkpoint_SD2_Config, ControlNet_Checkpoint_SD2_Config.get_tag()],
        Annotated[ControlNet_Checkpoint_SDXL_Config, ControlNet_Checkpoint_SDXL_Config.get_tag()],
        Annotated[ControlNet_Checkpoint_FLUX_Config, ControlNet_Checkpoint_FLUX_Config.get_tag()],
        # ControlNet - diffusers format
        Annotated[ControlNet_Diffusers_SD1_Config, ControlNet_Diffusers_SD1_Config.get_tag()],
        Annotated[ControlNet_Diffusers_SD2_Config, ControlNet_Diffusers_SD2_Config.get_tag()],
        Annotated[ControlNet_Diffusers_SDXL_Config, ControlNet_Diffusers_SDXL_Config.get_tag()],
        Annotated[ControlNet_Diffusers_FLUX_Config, ControlNet_Diffusers_FLUX_Config.get_tag()],
        # LoRA - LyCORIS format
        Annotated[LoRA_LyCORIS_SD1_Config, LoRA_LyCORIS_SD1_Config.get_tag()],
        Annotated[LoRA_LyCORIS_SD2_Config, LoRA_LyCORIS_SD2_Config.get_tag()],
        Annotated[LoRA_LyCORIS_SDXL_Config, LoRA_LyCORIS_SDXL_Config.get_tag()],
        Annotated[LoRA_LyCORIS_FLUX_Config, LoRA_LyCORIS_FLUX_Config.get_tag()],
        # LoRA - OMI format
        Annotated[LoRA_OMI_SDXL_Config, LoRA_OMI_SDXL_Config.get_tag()],
        Annotated[LoRA_OMI_FLUX_Config, LoRA_OMI_FLUX_Config.get_tag()],
        # LoRA - diffusers format
        Annotated[LoRA_Diffusers_SD1_Config, LoRA_Diffusers_SD1_Config.get_tag()],
        Annotated[LoRA_Diffusers_SD2_Config, LoRA_Diffusers_SD2_Config.get_tag()],
        Annotated[LoRA_Diffusers_SDXL_Config, LoRA_Diffusers_SDXL_Config.get_tag()],
        Annotated[LoRA_Diffusers_FLUX_Config, LoRA_Diffusers_FLUX_Config.get_tag()],
        # ControlLoRA - diffusers format
        Annotated[ControlLoRA_LyCORIS_FLUX_Config, ControlLoRA_LyCORIS_FLUX_Config.get_tag()],
        # T5 Encoder - all formats
        Annotated[T5Encoder_T5Encoder_Config, T5Encoder_T5Encoder_Config.get_tag()],
        Annotated[T5Encoder_BnBLLMint8_Config, T5Encoder_BnBLLMint8_Config.get_tag()],
        # TI - file format
        Annotated[TI_File_SD1_Config, TI_File_SD1_Config.get_tag()],
        Annotated[TI_File_SD2_Config, TI_File_SD2_Config.get_tag()],
        Annotated[TI_File_SDXL_Config, TI_File_SDXL_Config.get_tag()],
        # TI - folder format
        Annotated[TI_Folder_SD1_Config, TI_Folder_SD1_Config.get_tag()],
        Annotated[TI_Folder_SD2_Config, TI_Folder_SD2_Config.get_tag()],
        Annotated[TI_Folder_SDXL_Config, TI_Folder_SDXL_Config.get_tag()],
        # IP Adapter - InvokeAI format
        Annotated[IPAdapter_InvokeAI_SD1_Config, IPAdapter_InvokeAI_SD1_Config.get_tag()],
        Annotated[IPAdapter_InvokeAI_SD2_Config, IPAdapter_InvokeAI_SD2_Config.get_tag()],
        Annotated[IPAdapter_InvokeAI_SDXL_Config, IPAdapter_InvokeAI_SDXL_Config.get_tag()],
        # IP Adapter - checkpoint format
        Annotated[IPAdapter_Checkpoint_SD1_Config, IPAdapter_Checkpoint_SD1_Config.get_tag()],
        Annotated[IPAdapter_Checkpoint_SD2_Config, IPAdapter_Checkpoint_SD2_Config.get_tag()],
        Annotated[IPAdapter_Checkpoint_SDXL_Config, IPAdapter_Checkpoint_SDXL_Config.get_tag()],
        Annotated[IPAdapter_Checkpoint_FLUX_Config, IPAdapter_Checkpoint_FLUX_Config.get_tag()],
        # T2I Adapter - diffusers format
        Annotated[T2IAdapter_Diffusers_SD1_Config, T2IAdapter_Diffusers_SD1_Config.get_tag()],
        Annotated[T2IAdapter_Diffusers_SDXL_Config, T2IAdapter_Diffusers_SDXL_Config.get_tag()],
        # Misc models
        Annotated[Spandrel_Checkpoint_Config, Spandrel_Checkpoint_Config.get_tag()],
        Annotated[CLIPEmbed_Diffusers_G_Config, CLIPEmbed_Diffusers_G_Config.get_tag()],
        Annotated[CLIPEmbed_Diffusers_L_Config, CLIPEmbed_Diffusers_L_Config.get_tag()],
        Annotated[CLIPVision_Diffusers_Config, CLIPVision_Diffusers_Config.get_tag()],
        Annotated[SigLIP_Diffusers_Config, SigLIP_Diffusers_Config.get_tag()],
        Annotated[FLUXRedux_Checkpoint_Config, FLUXRedux_Checkpoint_Config.get_tag()],
        Annotated[LlavaOnevision_Diffusers_Config, LlavaOnevision_Diffusers_Config.get_tag()],
        # Main - external API
        Annotated[Main_ExternalAPI_ChatGPT4o_Config, Main_ExternalAPI_ChatGPT4o_Config.get_tag()],
        Annotated[Main_ExternalAPI_Gemini2_5_Config, Main_ExternalAPI_Gemini2_5_Config.get_tag()],
        Annotated[Main_ExternalAPI_Imagen3_Config, Main_ExternalAPI_Imagen3_Config.get_tag()],
        Annotated[Main_ExternalAPI_Imagen4_Config, Main_ExternalAPI_Imagen4_Config.get_tag()],
        Annotated[Main_ExternalAPI_FluxKontext_Config, Main_ExternalAPI_FluxKontext_Config.get_tag()],
        # Video - external API
        Annotated[Video_ExternalAPI_Veo3_Config, Video_ExternalAPI_Veo3_Config.get_tag()],
        Annotated[Video_ExternalAPI_Runway_Config, Video_ExternalAPI_Runway_Config.get_tag()],
        # Unknown model (fallback)
        Annotated[Unknown_Config, Unknown_Config.get_tag()],
    ],
    Discriminator(Config_Base.get_model_discriminator_value),
]

AnyModelConfigValidator = TypeAdapter[AnyModelConfig](AnyModelConfig)
"""Pydantic TypeAdapter for the AnyModelConfig union, used for parsing and validation.

If you need to parse/validate a dict or JSON into an AnyModelConfig, you should probably use
ModelConfigFactory.from_dict or ModelConfigFactory.from_json instead as they may implement
additional logic in the future.
"""


@dataclass
class ModelClassificationResult:
    """Result of attempting to classify a model on disk into a specific model config.

    Attributes:
        match: The best matching model config, or None if no match was found.
        results: A mapping of model config class names to either an instance of that class (if it matched)
            or an Exception (if it didn't match or an error occurred during matching).
    """

    config: AnyModelConfig | None
    details: dict[str, AnyModelConfig | Exception]

    @property
    def all_matches(self) -> list[AnyModelConfig]:
        """Returns a list of all matching model configs found."""
        return [r for r in self.details.values() if isinstance(r, Config_Base)]

    @property
    def match_count(self) -> int:
        """Returns the number of matching model configs found."""
        return len(self.all_matches)


class ModelConfigFactory:
    @staticmethod
    def from_dict(fields: dict[str, Any]) -> AnyModelConfig:
        """Return the appropriate config object from raw dict values."""
        model = AnyModelConfigValidator.validate_python(fields)
        return model

    @staticmethod
    def from_json(json: str | bytes | bytearray) -> AnyModelConfig:
        """Return the appropriate config object from json."""
        model = AnyModelConfigValidator.validate_json(json)
        return model

    @staticmethod
    def build_common_fields(
        mod: ModelOnDisk,
        override_fields: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Builds the common fields for all model configs.

        Args:
            mod: The model on disk to extract fields from.
            overrides: A optional dictionary of fields to override. These fields will take precedence over the values
                extracted from the model on disk.

        - Casts string fields to their Enum types.
        - Does not validate the fields against the model config schema.
        """

        _overrides: dict[str, Any] = override_fields or {}
        fields: dict[str, Any] = {}

        if "type" in _overrides:
            fields["type"] = ModelType(_overrides["type"])

        if "format" in _overrides:
            fields["format"] = ModelFormat(_overrides["format"])

        if "base" in _overrides:
            fields["base"] = BaseModelType(_overrides["base"])

        if "source_type" in _overrides:
            fields["source_type"] = ModelSourceType(_overrides["source_type"])

        if "variant" in _overrides:
            fields["variant"] = variant_type_adapter.validate_strings(_overrides["variant"])

        fields["path"] = mod.path.as_posix()
        fields["source"] = _overrides.get("source") or fields["path"]
        fields["source_type"] = _overrides.get("source_type") or ModelSourceType.Path
        fields["name"] = _overrides.get("name") or mod.name
        fields["hash"] = _overrides.get("hash") or mod.hash()
        fields["key"] = _overrides.get("key") or uuid_string()
        fields["description"] = _overrides.get("description")
        fields["file_size"] = _overrides.get("file_size") or mod.size()

        return fields

    @staticmethod
    def _validate_path_looks_like_model(path: Path) -> None:
        """Perform basic sanity checks to ensure a path looks like a model.

        This prevents wasting time trying to identify obviously non-model paths like
        home directories or downloads folders. Raises RuntimeError if the path doesn't
        pass basic checks.

        Args:
            path: The path to validate

        Raises:
            ValueError: If the path doesn't look like a model
        """
        if path.is_file():
            # For files, just check the extension
            if path.suffix.lower() not in _MODEL_EXTENSIONS:
                raise ValueError(
                    f"File extension {path.suffix} is not a recognized model format. "
                    f"Expected one of: {', '.join(sorted(_MODEL_EXTENSIONS))}"
                )
        else:
            # For directories, do a quick file count check with early exit
            total_files = 0
            # Ignore hidden files and directories
            paths_to_check = (p for p in path.rglob("*") if not p.name.startswith("."))
            for item in paths_to_check:
                if item.is_file():
                    total_files += 1
                    if total_files > _MAX_FILES_IN_MODEL_DIR:
                        raise ValueError(
                            f"Directory contains more than {_MAX_FILES_IN_MODEL_DIR} files. "
                            "This looks like a general-purpose directory rather than a model. "
                            "Please provide a path to a specific model file or model directory."
                        )

            # Check if it has config files at root (diffusers/transformers marker)
            has_root_config = any((path / config).exists() for config in _CONFIG_FILES)

            if has_root_config:
                # Has a config file, looks like a valid model directory
                return

            # Otherwise, search for model files within depth limit
            def find_model_files(current_path: Path, depth: int) -> bool:
                if depth > _MAX_SEARCH_DEPTH:
                    return False
                try:
                    for item in current_path.iterdir():
                        if item.is_file() and item.suffix.lower() in _MODEL_EXTENSIONS:
                            return True
                        elif item.is_dir() and find_model_files(item, depth + 1):
                            return True
                except PermissionError:
                    pass
                return False

            if not find_model_files(path, 0):
                raise ValueError(
                    f"No model files or config files found in directory {path}. "
                    f"Expected to find model files with extensions: {', '.join(sorted(_MODEL_EXTENSIONS))} "
                    f"or config files: {', '.join(sorted(_CONFIG_FILES))}"
                )

    @staticmethod
    def matches_sort_key(m: AnyModelConfig) -> int:
        """Sort key function to prioritize model config matches in case of multiple matches."""

        # It is possible that we have multiple matches. We need to prioritize them.

        # Known cases where multiple matches can occur:
        # - SD main models can look like a LoRA when they have merged in LoRA weights. Prefer the main model.
        # - SD main models in diffusers format can look like a CLIP Embed; they have a text_encoder folder with
        #   a config.json file. Prefer the main model.

        # Given the above cases, we can prioritize the matches by type. If we find more cases, we may need a more
        # sophisticated approach.
        match m.type:
            case ModelType.Main:
                return 0
            case ModelType.LoRA:
                return 1
            case ModelType.CLIPEmbed:
                return 2
            case _:
                return 3

    @staticmethod
    def from_model_on_disk(
        mod: str | Path | ModelOnDisk,
        override_fields: dict[str, Any] | None = None,
        hash_algo: HASHING_ALGORITHMS = "blake3_single",
        allow_unknown: bool = True,
    ) -> ModelClassificationResult:
        """Classify a model on disk and return the best matching model config.

        Args:
            mod: The model on disk to classify. Can be a path (str or Path) or a ModelOnDisk instance.
            override_fields: Optional dictionary of fields to override. These fields will take precedence
                over the values extracted from the model on disk, but this cannot force a match if the
                model on disk doesn't actually match the config class.
            hash_algo: The hashing algorithm to use when computing the model hash if needed.

        Returns:
            A ModelClassificationResult containing the best matching model config (or None if no match)
            and a mapping of all attempted model config classes to either an instance of that class (if it matched)
            or an Exception (if it didn't match or an error occurred during matching).

        Raises:
            ValueError: If the provided path doesn't look like a model.
        """
        if isinstance(mod, Path | str):
            mod = ModelOnDisk(Path(mod), hash_algo)

        # Perform basic sanity checks before attempting any config matching
        # This rejects obviously non-model paths early, saving time
        ModelConfigFactory._validate_path_looks_like_model(mod.path)

        # We will always need these fields to build any model config.
        fields = ModelConfigFactory.build_common_fields(mod, override_fields)

        # Store results as a mapping of config class to either an instance of that class or an exception
        # that was raised when trying to build it.
        details: dict[str, AnyModelConfig | Exception] = {}

        # Try to build an instance of each model config class that uses the classify API.
        # Each class will either return an instance of itself or raise NotAMatch if it doesn't match.
        # Other exceptions may be raised if something unexpected happens during matching or building.
        for candidate_class in filter(lambda x: x is not Unknown_Config, Config_Base.CONFIG_CLASSES):
            candidate_name = candidate_class.__name__
            try:
                # Technically, from_model_on_disk returns a Config_Base, but in practice it will always be a member of
                # the AnyModelConfig union.
                details[candidate_name] = candidate_class.from_model_on_disk(mod, fields)  # type: ignore
            except NotAMatchError as e:
                # This means the model didn't match this config class. It's not an error, just no match.
                details[candidate_name] = e
            except ValidationError as e:
                # This means the model matched, but we couldn't create the pydantic model instance for the config.
                # Maybe invalid overrides were provided?
                details[candidate_name] = e
            except Exception as e:
                # Some other unexpected error occurred. Store the exception for reporting later.
                details[candidate_name] = e

        # Extract just the successful matches
        matches = [r for r in details.values() if isinstance(r, Config_Base)]

        if not matches:
            if not allow_unknown:
                # No matches and we are not allowed to fall back to Unknown_Config
                return ModelClassificationResult(config=None, details=details)
            else:
                # Fall back to Unknown_Config
                # This should always succeed as Unknown_Config.from_model_on_disk never raises NotAMatch
                config = Unknown_Config.from_model_on_disk(mod, fields)
                details[Unknown_Config.__name__] = config
                return ModelClassificationResult(config=config, details=details)

        matches.sort(key=ModelConfigFactory.matches_sort_key)
        config = matches[0]

        # Now do any post-processing needed for specific model types/bases/etc.
        match config.type:
            case ModelType.Main:
                config.default_settings = MainModelDefaultSettings.from_base(config.base)
            case ModelType.ControlNet | ModelType.T2IAdapter | ModelType.ControlLoRa:
                config.default_settings = ControlAdapterDefaultSettings.from_model_name(config.name)
            case ModelType.LoRA:
                config.default_settings = LoraModelDefaultSettings()
            case _:
                pass

        return ModelClassificationResult(config=config, details=details)


MODEL_NAME_TO_PREPROCESSOR = {
    "canny": "canny_image_processor",
    "mlsd": "mlsd_image_processor",
    "depth": "depth_anything_image_processor",
    "bae": "normalbae_image_processor",
    "normal": "normalbae_image_processor",
    "sketch": "pidi_image_processor",
    "scribble": "lineart_image_processor",
    "lineart anime": "lineart_anime_image_processor",
    "lineart_anime": "lineart_anime_image_processor",
    "lineart": "lineart_image_processor",
    "soft": "hed_image_processor",
    "softedge": "hed_image_processor",
    "hed": "hed_image_processor",
    "shuffle": "content_shuffle_image_processor",
    "pose": "dw_openpose_image_processor",
    "mediapipe": "mediapipe_face_processor",
    "pidi": "pidi_image_processor",
    "zoe": "zoe_depth_image_processor",
    "color": "color_map_image_processor",
}



from typing import (
    Literal,
    Self,
)

from pydantic import Field
from typing_extensions import Any

from invokeai.backend.model_manager.configs.base import Config_Base, Diffusers_Config_Base
from invokeai.backend.model_manager.configs.identification_utils import (
    NotAMatchError,
    get_config_dict_or_raise,
    raise_for_class_name,
    raise_for_override_fields,
    raise_if_not_dir,
)
from invokeai.backend.model_manager.model_on_disk import ModelOnDisk
from invokeai.backend.model_manager.taxonomy import (
    BaseModelType,
    ClipVariantType,
    ModelFormat,
    ModelType,
)


def get_clip_variant_type_from_config(config: dict[str, Any]) -> ClipVariantType | None:
    try:
        hidden_size = config.get("hidden_size")
        match hidden_size:
            case 1280:
                return ClipVariantType.G
            case 768:
                return ClipVariantType.L
            case _:
                return None
    except Exception:
        return None


class CLIPEmbed_Diffusers_Config_Base(Diffusers_Config_Base):
    base: Literal[BaseModelType.Any] = Field(default=BaseModelType.Any)
    type: Literal[ModelType.CLIPEmbed] = Field(default=ModelType.CLIPEmbed)
    format: Literal[ModelFormat.Diffusers] = Field(default=ModelFormat.Diffusers)

    @classmethod
    def from_model_on_disk(cls, mod: ModelOnDisk, override_fields: dict[str, Any]) -> Self:
        raise_if_not_dir(mod)

        raise_for_override_fields(cls, override_fields)

        raise_for_class_name(
            {
                mod.path / "config.json",
                mod.path / "text_encoder" / "config.json",
            },
            {
                "CLIPModel",
                "CLIPTextModel",
                "CLIPTextModelWithProjection",
            },
        )

        cls._validate_variant(mod)

        return cls(**override_fields)

    @classmethod
    def _validate_variant(cls, mod: ModelOnDisk) -> None:
        """Raise `NotAMatch` if the model variant does not match this config class."""
        expected_variant = cls.model_fields["variant"].default
        config = get_config_dict_or_raise(
            {
                mod.path / "config.json",
                mod.path / "text_encoder" / "config.json",
            },
        )
        recognized_variant = get_clip_variant_type_from_config(config)

        if recognized_variant is None:
            raise NotAMatchError("unable to determine CLIP variant from config")

        if expected_variant is not recognized_variant:
            raise NotAMatchError(f"variant is {recognized_variant}, not {expected_variant}")


class CLIPEmbed_Diffusers_G_Config(CLIPEmbed_Diffusers_Config_Base, Config_Base):
    variant: Literal[ClipVariantType.G] = Field(default=ClipVariantType.G)


class CLIPEmbed_Diffusers_L_Config(CLIPEmbed_Diffusers_Config_Base, Config_Base):
    variant: Literal[ClipVariantType.L] = Field(default=ClipVariantType.L)



stable_diffusion_1_lora = "stable-diffusion-v1/lora"
stable_diffusion_1_inpainting_lora = "stable-diffusion-v1-inpainting/lora"

stable_diffusion_2_512_lora = "stable-diffusion-v2-512/lora"
stable_diffusion_2_768_v_lora = "stable-diffusion-v2-768-v/lora"
stable_diffusion_2_depth_lora = "stable-diffusion-v2-depth/lora"
stable_diffusion_2_inpainting_lora = "stable-diffusion-v2-inpainting/lora"

stable_diffusion_3_medium_lora = "stable-diffusion-v3-medium/lora"
stable_diffusion_35_medium_lora = "stable-diffusion-v3.5-medium/lora"
stable_diffusion_35_large_lora = "stable-diffusion-v3.5-large/lora"

stable_diffusion_xl_1_lora = "stable-diffusion-xl-v1-base/lora"
stable_diffusion_xl_1_inpainting_lora = "stable-diffusion-xl-v1-base-inpainting/lora"

wuerstchen_2_lora = "wuerstchen-v2-prior/lora"
stable_cascade_1_stage_a_lora = "stable-cascade-v1-stage-a/lora"
stable_cascade_1_stage_b_lora = "stable-cascade-v1-stage-b/lora"
stable_cascade_1_stage_c_lora = "stable-cascade-v1-stage-c/lora"

pixart_alpha_lora = "pixart-alpha/lora"
pixart_sigma_lora = "pixart-sigma/lora"

flux_dev_1_lora = "Flux.1-dev/lora"
flux_fill_dev_1_lora = "Flux.1-fill-dev/lora"

sana_lora = "sana/lora"

hunyuan_video_lora = "hunyuan-video/lora"

hi_dream_i1_lora = "hidream-i1/lora"



from dataclasses import dataclass

from invokeai.backend.model_manager.configs.factory import AnyModelConfig
from invokeai.backend.model_manager.taxonomy import (
    BaseModelType,
    ModelType,
    ModelVariantType,
    SchedulerPredictionType,
)


@dataclass(frozen=True)
class LegacyConfigKey:
    type: ModelType
    base: BaseModelType
    variant: ModelVariantType | None = None
    pred: SchedulerPredictionType | None = None

    @classmethod
    def from_model_config(cls, config: AnyModelConfig) -> "LegacyConfigKey":
        variant = getattr(config, "variant", None)
        pred = getattr(config, "prediction_type", None)
        return cls(type=config.type, base=config.base, variant=variant, pred=pred)


LEGACY_CONFIG_MAP: dict[LegacyConfigKey, str] = {
    LegacyConfigKey(
        ModelType.Main,
        BaseModelType.StableDiffusion1,
        ModelVariantType.Normal,
        SchedulerPredictionType.Epsilon,
    ): "stable-diffusion/v1-inference.yaml",
    LegacyConfigKey(
        ModelType.Main,
        BaseModelType.StableDiffusion1,
        ModelVariantType.Normal,
        SchedulerPredictionType.VPrediction,
    ): "stable-diffusion/v1-inference-v.yaml",
    LegacyConfigKey(
        ModelType.Main,
        BaseModelType.StableDiffusion1,
        ModelVariantType.Inpaint,
    ): "stable-diffusion/v1-inpainting-inference.yaml",
    LegacyConfigKey(
        ModelType.Main,
        BaseModelType.StableDiffusion2,
        ModelVariantType.Normal,
        SchedulerPredictionType.Epsilon,
    ): "stable-diffusion/v2-inference.yaml",
    LegacyConfigKey(
        ModelType.Main,
        BaseModelType.StableDiffusion2,
        ModelVariantType.Normal,
        SchedulerPredictionType.VPrediction,
    ): "stable-diffusion/v2-inference-v.yaml",
    LegacyConfigKey(
        ModelType.Main,
        BaseModelType.StableDiffusion2,
        ModelVariantType.Inpaint,
        SchedulerPredictionType.Epsilon,
    ): "stable-diffusion/v2-inpainting-inference.yaml",
    LegacyConfigKey(
        ModelType.Main,
        BaseModelType.StableDiffusion2,
        ModelVariantType.Inpaint,
        SchedulerPredictionType.VPrediction,
    ): "stable-diffusion/v2-inpainting-inference-v.yaml",
    LegacyConfigKey(
        ModelType.Main,
        BaseModelType.StableDiffusion2,
        ModelVariantType.Depth,
    ): "stable-diffusion/v2-midas-inference.yaml",
    LegacyConfigKey(
        ModelType.Main,
        BaseModelType.StableDiffusionXL,
        ModelVariantType.Normal,
    ): "stable-diffusion/sd_xl_base.yaml",
    LegacyConfigKey(
        ModelType.Main,
        BaseModelType.StableDiffusionXL,
        ModelVariantType.Inpaint,
    ): "stable-diffusion/sd_xl_inpaint.yaml",
    LegacyConfigKey(
        ModelType.Main,
        BaseModelType.StableDiffusionXLRefiner,
        ModelVariantType.Normal,
    ): "stable-diffusion/sd_xl_refiner.yaml",
    LegacyConfigKey(ModelType.ControlNet, BaseModelType.StableDiffusion1): "controlnet/cldm_v15.yaml",
    LegacyConfigKey(ModelType.ControlNet, BaseModelType.StableDiffusion2): "controlnet/cldm_v21.yaml",
    LegacyConfigKey(ModelType.VAE, BaseModelType.StableDiffusion1): "stable-diffusion/v1-inference.yaml",
    LegacyConfigKey(ModelType.VAE, BaseModelType.StableDiffusion2): "stable-diffusion/v2-inference.yaml",
    LegacyConfigKey(ModelType.VAE, BaseModelType.StableDiffusionXL): "stable-diffusion/sd_xl_base.yaml",
}



from typing import (
    Literal,
    Self,
)

from pydantic import Field
from typing_extensions import Any

from invokeai.backend.model_manager.configs.base import Config_Base, Diffusers_Config_Base
from invokeai.backend.model_manager.configs.controlnet import ControlAdapterDefaultSettings
from invokeai.backend.model_manager.configs.identification_utils import (
    NotAMatchError,
    common_config_paths,
    get_config_dict_or_raise,
    raise_for_class_name,
    raise_for_override_fields,
    raise_if_not_dir,
)
from invokeai.backend.model_manager.model_on_disk import ModelOnDisk
from invokeai.backend.model_manager.taxonomy import (
    BaseModelType,
    ModelFormat,
    ModelType,
)


class T2IAdapter_Diffusers_Config_Base(Diffusers_Config_Base):
    """Model config for T2I."""

    type: Literal[ModelType.T2IAdapter] = Field(default=ModelType.T2IAdapter)
    format: Literal[ModelFormat.Diffusers] = Field(default=ModelFormat.Diffusers)
    default_settings: ControlAdapterDefaultSettings | None = Field(None)

    @classmethod
    def from_model_on_disk(cls, mod: ModelOnDisk, override_fields: dict[str, Any]) -> Self:
        raise_if_not_dir(mod)

        raise_for_override_fields(cls, override_fields)

        raise_for_class_name(
            common_config_paths(mod.path),
            {
                "T2IAdapter",
            },
        )

        cls._validate_base(mod)

        return cls(**override_fields)

    @classmethod
    def _validate_base(cls, mod: ModelOnDisk) -> None:
        """Raise `NotAMatch` if the model base does not match this config class."""
        expected_base = cls.model_fields["base"].default
        recognized_base = cls._get_base_or_raise(mod)
        if expected_base is not recognized_base:
            raise NotAMatchError(f"base is {recognized_base}, not {expected_base}")

    @classmethod
    def _get_base_or_raise(cls, mod: ModelOnDisk) -> BaseModelType:
        config_dict = get_config_dict_or_raise(common_config_paths(mod.path))

        adapter_type = config_dict.get("adapter_type")

        match adapter_type:
            case "full_adapter_xl":
                return BaseModelType.StableDiffusionXL
            case "full_adapter" | "light_adapter":
                return BaseModelType.StableDiffusion1
            case _:
                raise NotAMatchError(f"unrecognized adapter_type '{adapter_type}'")


class T2IAdapter_Diffusers_SD1_Config(T2IAdapter_Diffusers_Config_Base, Config_Base):
    base: Literal[BaseModelType.StableDiffusion1] = Field(default=BaseModelType.StableDiffusion1)


class T2IAdapter_Diffusers_SDXL_Config(T2IAdapter_Diffusers_Config_Base, Config_Base):
    base: Literal[BaseModelType.StableDiffusionXL] = Field(default=BaseModelType.StableDiffusionXL)



import sqlite3
from typing import Optional, cast

from invokeai.app.services.board_image_records.board_image_records_base import BoardImageRecordStorageBase
from invokeai.app.services.image_records.image_records_common import (
    ASSETS_CATEGORIES,
    IMAGE_CATEGORIES,
    ImageCategory,
    ImageRecord,
    deserialize_image_record,
)
from invokeai.app.services.shared.pagination import OffsetPaginatedResults
from invokeai.app.services.shared.sqlite.sqlite_database import SqliteDatabase


class SqliteBoardImageRecordStorage(BoardImageRecordStorageBase):
    def __init__(self, db: SqliteDatabase) -> None:
        super().__init__()
        self._db = db

    def add_image_to_board(
        self,
        board_id: str,
        image_name: str,
    ) -> None:
        with self._db.transaction() as cursor:
            cursor.execute(
                """--sql
                INSERT INTO board_images (board_id, image_name)
                VALUES (?, ?)
                ON CONFLICT (image_name) DO UPDATE SET board_id = ?;
                """,
                (board_id, image_name, board_id),
            )

    def remove_image_from_board(
        self,
        image_name: str,
    ) -> None:
        with self._db.transaction() as cursor:
            cursor.execute(
                """--sql
                DELETE FROM board_images
                WHERE image_name = ?;
                """,
                (image_name,),
            )

    def get_images_for_board(
        self,
        board_id: str,
        offset: int = 0,
        limit: int = 10,
    ) -> OffsetPaginatedResults[ImageRecord]:
        with self._db.transaction() as cursor:
            cursor.execute(
                """--sql
                SELECT images.*
                FROM board_images
                INNER JOIN images ON board_images.image_name = images.image_name
                WHERE board_images.board_id = ?
                ORDER BY board_images.updated_at DESC;
                """,
                (board_id,),
            )
            result = cast(list[sqlite3.Row], cursor.fetchall())
            images = [deserialize_image_record(dict(r)) for r in result]

            cursor.execute(
                """--sql
                SELECT COUNT(*) FROM images WHERE 1=1;
                """
            )
            count = cast(int, cursor.fetchone()[0])

        return OffsetPaginatedResults(items=images, offset=offset, limit=limit, total=count)

    def get_all_board_image_names_for_board(
        self,
        board_id: str,
        categories: list[ImageCategory] | None,
        is_intermediate: bool | None,
    ) -> list[str]:
        with self._db.transaction() as cursor:
            params: list[str | bool] = []

            # Base query is a join between images and board_images
            stmt = """
                    SELECT images.image_name
                    FROM images
                    LEFT JOIN board_images ON board_images.image_name = images.image_name
                    WHERE 1=1
                    """

            # Handle board_id filter
            if board_id == "none":
                stmt += """--sql
                    AND board_images.board_id IS NULL
                    """
            else:
                stmt += """--sql
                    AND board_images.board_id = ?
                    """
                params.append(board_id)

            # Add the category filter
            if categories is not None:
                # Convert the enum values to unique list of strings
                category_strings = [c.value for c in set(categories)]
                # Create the correct length of placeholders
                placeholders = ",".join("?" * len(category_strings))
                stmt += f"""--sql
                    AND images.image_category IN ( {placeholders} )
                    """

                # Unpack the included categories into the query params
                for c in category_strings:
                    params.append(c)

            # Add the is_intermediate filter
            if is_intermediate is not None:
                stmt += """--sql
                    AND images.is_intermediate = ?
                    """
                params.append(is_intermediate)

            # Put a ring on it
            stmt += ";"

            cursor.execute(stmt, params)

            result = cast(list[sqlite3.Row], cursor.fetchall())
        image_names = [r[0] for r in result]
        return image_names

    def get_board_for_image(
        self,
        image_name: str,
    ) -> Optional[str]:
        with self._db.transaction() as cursor:
            cursor.execute(
                """--sql
                    SELECT board_id
                    FROM board_images
                    WHERE image_name = ?;
                    """,
                (image_name,),
            )
            result = cursor.fetchone()
        if result is None:
            return None
        return cast(str, result[0])

    def get_image_count_for_board(self, board_id: str) -> int:
        with self._db.transaction() as cursor:
            # Convert the enum values to unique list of strings
            category_strings = [c.value for c in set(IMAGE_CATEGORIES)]
            # Create the correct length of placeholders
            placeholders = ",".join("?" * len(category_strings))
            cursor.execute(
                f"""--sql
                    SELECT COUNT(*)
                    FROM board_images
                    INNER JOIN images ON board_images.image_name = images.image_name
                    WHERE images.is_intermediate = FALSE AND images.image_category IN ( {placeholders} )
                    AND board_images.board_id = ?;
                    """,
                (*category_strings, board_id),
            )
            count = cast(int, cursor.fetchone()[0])
        return count

    def get_asset_count_for_board(self, board_id: str) -> int:
        with self._db.transaction() as cursor:
            # Convert the enum values to unique list of strings
            category_strings = [c.value for c in set(ASSETS_CATEGORIES)]
            # Create the correct length of placeholders
            placeholders = ",".join("?" * len(category_strings))
            cursor.execute(
                f"""--sql
                    SELECT COUNT(*)
                    FROM board_images
                    INNER JOIN images ON board_images.image_name = images.image_name
                    WHERE images.is_intermediate = FALSE AND images.image_category IN ( {placeholders} )
                    AND board_images.board_id = ?;
                    """,
                (*category_strings, board_id),
            )
            count = cast(int, cursor.fetchone()[0])
        return count



import json
import sqlite3
from logging import Logger
from pathlib import Path
from typing import NamedTuple

from pydantic import ValidationError

from invokeai.app.services.config import InvokeAIAppConfig
from invokeai.app.services.shared.sqlite_migrator.sqlite_migrator_common import Migration
from invokeai.backend.model_manager.configs.factory import AnyModelConfigValidator


class NormalizeResult(NamedTuple):
    new_relative_path: str | None
    rollback_ops: list[tuple[Path, Path]]


class Migration24Callback:
    def __init__(self, app_config: InvokeAIAppConfig, logger: Logger) -> None:
        self._app_config = app_config
        self._logger = logger
        self._models_dir = app_config.models_path.resolve()

    def __call__(self, cursor: sqlite3.Cursor) -> None:
        # Grab all model records
        cursor.execute("SELECT id, config FROM models;")
        rows = cursor.fetchall()

        for model_id, config_json in rows:
            try:
                config = AnyModelConfigValidator.validate_json(config_json)
            except ValidationError:
                # This could happen if the config schema changed in a way that makes old configs invalid. Unlikely
                # for users, more likely for devs testing out migration paths.
                self._logger.warning("Skipping model %s: invalid config schema", model_id)
                continue
            except json.JSONDecodeError:
                # This should never happen, as we use pydantic to serialize the config to JSON.
                self._logger.warning("Skipping model %s: invalid config JSON", model_id)
                continue

            # We'll use a savepoint so we can roll back the database update if something goes wrong, and a simple
            # rollback of file operations if needed.
            cursor.execute("SAVEPOINT migrate_model")
            try:
                new_relative_path, rollback_ops = self._normalize_model_storage(
                    key=config.key,
                    path_value=config.path,
                )
            except Exception as err:
                self._logger.error("Error normalizing model %s: %s", config.key, err)
                cursor.execute("ROLLBACK TO SAVEPOINT migrate_model")
                cursor.execute("RELEASE SAVEPOINT migrate_model")
                continue

            if new_relative_path is None:
                cursor.execute("RELEASE SAVEPOINT migrate_model")
                continue

            config.path = new_relative_path
            try:
                cursor.execute(
                    "UPDATE models SET config = ? WHERE id = ?;",
                    (config.model_dump_json(), model_id),
                )
            except Exception as err:
                self._logger.error("Database update failed for model %s: %s", config.key, err)
                cursor.execute("ROLLBACK TO SAVEPOINT migrate_model")
                cursor.execute("RELEASE SAVEPOINT migrate_model")
                self._rollback_file_ops(rollback_ops)
                continue

            cursor.execute("RELEASE SAVEPOINT migrate_model")

        self._prune_empty_directories()

    def _normalize_model_storage(self, key: str, path_value: str) -> NormalizeResult:
        models_dir = self._models_dir
        stored_path = Path(path_value)

        relative_path: Path | None
        if stored_path.is_absolute():
            # If the stored path is absolute, we need to check if it's inside the models directory, which means it is
            # an Invoke-managed model. If it's outside, it is user-managed we leave it alone.
            try:
                relative_path = stored_path.resolve().relative_to(models_dir)
            except ValueError:
                self._logger.info("Leaving user-managed model %s at %s", key, stored_path)
                return NormalizeResult(new_relative_path=None, rollback_ops=[])
        else:
            # Relative paths are always relative to the models directory and thus Invoke-managed.
            relative_path = stored_path

        # If the relative path is empty, assume something is wrong. Warn and skip.
        if not relative_path.parts:
            self._logger.warning("Skipping model %s: empty relative path", key)
            return NormalizeResult(new_relative_path=None, rollback_ops=[])

        # Sanity check: the path is relative. It should be present in the models directory.
        absolute_path = (models_dir / relative_path).resolve()
        if not absolute_path.exists():
            self._logger.warning(
                "Skipping model %s: expected model files at %s but nothing was found",
                key,
                absolute_path,
            )
            return NormalizeResult(new_relative_path=None, rollback_ops=[])

        if relative_path.parts[0] == key:
            # Already normalized. Still ensure the stored path is relative.
            normalized_path = relative_path.as_posix()
            # If the stored path is already the normalized path, no change is needed.
            new_relative_path = normalized_path if stored_path.as_posix() != normalized_path else None
            return NormalizeResult(new_relative_path=new_relative_path, rollback_ops=[])

        # We'll store the file operations we perform so we can roll them back if needed.
        rollback_ops: list[tuple[Path, Path]] = []

        # Destination directory is models_dir/<key> - a flat directory structure.
        destination_dir = models_dir / key

        try:
            if absolute_path.is_file():
                destination_dir.mkdir(parents=True, exist_ok=True)
                dest_file = destination_dir / absolute_path.name
                # This really shouldn't happen.
                if dest_file.exists():
                    self._logger.warning(
                        "Destination for model %s already exists at %s; skipping move",
                        key,
                        dest_file,
                    )
                    return NormalizeResult(new_relative_path=None, rollback_ops=[])

                self._logger.info("Moving model file %s -> %s", absolute_path, dest_file)

                # `Path.rename()` effectively moves the file or directory.
                absolute_path.rename(dest_file)
                rollback_ops.append((dest_file, absolute_path))

                return NormalizeResult(
                    new_relative_path=(Path(key) / dest_file.name).as_posix(),
                    rollback_ops=rollback_ops,
                )

            if absolute_path.is_dir():
                dest_path = destination_dir
                # This really shouldn't happen.
                if dest_path.exists():
                    self._logger.warning(
                        "Destination directory %s already exists for model %s; skipping",
                        dest_path,
                        key,
                    )
                    return NormalizeResult(new_relative_path=None, rollback_ops=[])

                self._logger.info("Moving model directory %s -> %s", absolute_path, dest_path)

                # `Path.rename()` effectively moves the file or directory.
                absolute_path.rename(dest_path)
                rollback_ops.append((dest_path, absolute_path))

                return NormalizeResult(
                    new_relative_path=Path(key).as_posix(),
                    rollback_ops=rollback_ops,
                )

            # Maybe a broken symlink or something else weird?
            self._logger.warning("Skipping model %s: path %s is neither a file nor directory", key, absolute_path)
            return NormalizeResult(new_relative_path=None, rollback_ops=[])
        except Exception:
            self._rollback_file_ops(rollback_ops)
            raise

    def _rollback_file_ops(self, rollback_ops: list[tuple[Path, Path]]) -> None:
        # This is a super-simple rollback that just reverses the move operations we performed.
        for source, destination in reversed(rollback_ops):
            try:
                if source.exists():
                    source.rename(destination)
            except Exception as err:
                self._logger.error("Failed to rollback move %s -> %s: %s", source, destination, err)

    def _prune_empty_directories(self) -> None:
        # These directories are system directories we want to keep even if empty. Technically, the app should not
        # have any problems if these are removed, creating them as needed, but it's cleaner to just leave them alone.
        keep_names = {"model_images", ".download_cache"}
        keep_dirs = {self._models_dir / name for name in keep_names}
        removed_dirs: set[Path] = set()

        # Walk the models directory tree from the bottom up, removing empty directories. We sort by path length
        # descending to ensure we visit children before parents.
        for directory in sorted(self._models_dir.rglob("*"), key=lambda p: len(p.parts), reverse=True):
            if not directory.is_dir():
                continue
            if directory == self._models_dir:
                continue
            if any(directory == keep or keep in directory.parents for keep in keep_dirs):
                continue

            try:
                next(directory.iterdir())
            except StopIteration:
                try:
                    directory.rmdir()
                    removed_dirs.add(directory)
                    self._logger.debug("Removed empty directory %s", directory)
                except OSError:
                    # Directory not empty (or some other error) - bail out.
                    self._logger.warning("Failed to prune directory %s - not empty?", directory)
                    continue
            except OSError:
                continue

        self._logger.info("Pruned %d empty directories under %s", len(removed_dirs), self._models_dir)


def build_migration_24(app_config: InvokeAIAppConfig, logger: Logger) -> Migration:
    """Builds the migration object for migrating from version 23 to version 24.

    This migration normalizes on-disk model storage so that each model lives within
    a directory named by its key inside the Invoke-managed models directory, and
    updates database records to reference the new relative paths.

    This migration behaves a bit differently than others. Because it involves FS operations, if we rolled the
    DB back on any failure, we could leave the FS out of sync with the DB. Instead, we use savepoints
    to roll back individual model updates on failure, and we roll back any FS operations we performed
    for that model.

    If a model cannot be migrated for any reason (invalid config, missing files, FS errors, DB errors), we log a
    warning and skip it, leaving it in its original state and location. The model will still work, but it will be in
    the "wrong" location on disk.
    """

    return Migration(
        from_version=23,
        to_version=24,
        callback=Migration24Callback(app_config=app_config, logger=logger),
    )



from copy import deepcopy
from typing import Any, Literal, Self

from pydantic import Field

from invokeai.app.services.config.config_default import get_config
from invokeai.backend.model_manager.configs.base import Config_Base
from invokeai.backend.model_manager.model_on_disk import ModelOnDisk
from invokeai.backend.model_manager.taxonomy import (
    BaseModelType,
    ModelFormat,
    ModelType,
)

app_config = get_config()


class Unknown_Config(Config_Base):
    """Model config for unknown models, used as a fallback when we cannot positively identify a model."""

    base: Literal[BaseModelType.Unknown] = Field(default=BaseModelType.Unknown)
    type: Literal[ModelType.Unknown] = Field(default=ModelType.Unknown)
    format: Literal[ModelFormat.Unknown] = Field(default=ModelFormat.Unknown)

    @classmethod
    def from_model_on_disk(cls, mod: ModelOnDisk, override_fields: dict[str, Any]) -> Self:
        """Create an Unknown_Config for models that couldn't be positively identified.

        Note: Basic path validation (file extensions, directory structure) is already
        performed by ModelConfigFactory before this method is called.
        """

        cloned_override_fields = deepcopy(override_fields)
        cloned_override_fields.pop("base", None)
        cloned_override_fields.pop("type", None)
        cloned_override_fields.pop("format", None)

        return cls(
            **cloned_override_fields,
            # Override the type/format/base to ensure it's marked as unknown.
            base=BaseModelType.Unknown,
            type=ModelType.Unknown,
            format=ModelFormat.Unknown,
        )



from typing import (
    Literal,
    Self,
)

from pydantic import Field
from typing_extensions import Any

from invokeai.backend.flux.redux.flux_redux_state_dict_utils import is_state_dict_likely_flux_redux
from invokeai.backend.model_manager.configs.base import Config_Base
from invokeai.backend.model_manager.configs.identification_utils import (
    NotAMatchError,
    raise_for_override_fields,
    raise_if_not_file,
)
from invokeai.backend.model_manager.model_on_disk import ModelOnDisk
from invokeai.backend.model_manager.taxonomy import (
    BaseModelType,
    ModelFormat,
    ModelType,
)


class FLUXRedux_Checkpoint_Config(Config_Base):
    """Model config for FLUX Tools Redux model."""

    type: Literal[ModelType.FluxRedux] = Field(default=ModelType.FluxRedux)
    format: Literal[ModelFormat.Checkpoint] = Field(default=ModelFormat.Checkpoint)
    base: Literal[BaseModelType.Flux] = Field(default=BaseModelType.Flux)

    @classmethod
    def from_model_on_disk(cls, mod: ModelOnDisk, override_fields: dict[str, Any]) -> Self:
        raise_if_not_file(mod)

        raise_for_override_fields(cls, override_fields)

        if not is_state_dict_likely_flux_redux(mod.load_state_dict()):
            raise NotAMatchError("model does not match FLUX Tools Redux heuristics")

        return cls(**override_fields)



from typing import Optional

from fastapi import Body, HTTPException, Path, Query
from fastapi.routing import APIRouter

from invokeai.app.services.shared.pagination import OffsetPaginatedResults
from invokeai.app.services.shared.sqlite.sqlite_common import SQLiteDirection
from invokeai.app.services.videos_common import (
    DeleteVideosResult,
    StarredVideosResult,
    UnstarredVideosResult,
    VideoDTO,
    VideoIdsResult,
    VideoRecordChanges,
)

videos_router = APIRouter(prefix="/v1/videos", tags=["videos"])


@videos_router.patch(
    "/i/{video_id}",
    operation_id="update_video",
    response_model=VideoDTO,
)
async def update_video(
    video_id: str = Path(description="The id of the video to update"),
    video_changes: VideoRecordChanges = Body(description="The changes to apply to the video"),
) -> VideoDTO:
    """Updates a video"""

    raise HTTPException(status_code=501, detail="Not implemented")


@videos_router.get(
    "/i/{video_id}",
    operation_id="get_video_dto",
    response_model=VideoDTO,
)
async def get_video_dto(
    video_id: str = Path(description="The id of the video to get"),
) -> VideoDTO:
    """Gets a video's DTO"""

    raise HTTPException(status_code=501, detail="Not implemented")


@videos_router.post("/delete", operation_id="delete_videos_from_list", response_model=DeleteVideosResult)
async def delete_videos_from_list(
    video_ids: list[str] = Body(description="The list of ids of videos to delete", embed=True),
) -> DeleteVideosResult:
    raise HTTPException(status_code=501, detail="Not implemented")


@videos_router.post("/star", operation_id="star_videos_in_list", response_model=StarredVideosResult)
async def star_videos_in_list(
    video_ids: list[str] = Body(description="The list of ids of videos to star", embed=True),
) -> StarredVideosResult:
    raise HTTPException(status_code=501, detail="Not implemented")


@videos_router.post("/unstar", operation_id="unstar_videos_in_list", response_model=UnstarredVideosResult)
async def unstar_videos_in_list(
    video_ids: list[str] = Body(description="The list of ids of videos to unstar", embed=True),
) -> UnstarredVideosResult:
    raise HTTPException(status_code=501, detail="Not implemented")


@videos_router.delete("/uncategorized", operation_id="delete_uncategorized_videos", response_model=DeleteVideosResult)
async def delete_uncategorized_videos() -> DeleteVideosResult:
    """Deletes all videos that are uncategorized"""

    raise HTTPException(status_code=501, detail="Not implemented")


@videos_router.get("/", operation_id="list_video_dtos", response_model=OffsetPaginatedResults[VideoDTO])
async def list_video_dtos(
    is_intermediate: Optional[bool] = Query(default=None, description="Whether to list intermediate videos."),
    board_id: Optional[str] = Query(
        default=None,
        description="The board id to filter by. Use 'none' to find videos without a board.",
    ),
    offset: int = Query(default=0, description="The page offset"),
    limit: int = Query(default=10, description="The number of videos per page"),
    order_dir: SQLiteDirection = Query(default=SQLiteDirection.Descending, description="The order of sort"),
    starred_first: bool = Query(default=True, description="Whether to sort by starred videos first"),
    search_term: Optional[str] = Query(default=None, description="The term to search for"),
) -> OffsetPaginatedResults[VideoDTO]:
    """Lists video DTOs"""

    raise HTTPException(status_code=501, detail="Not implemented")


@videos_router.get("/ids", operation_id="get_video_ids")
async def get_video_ids(
    is_intermediate: Optional[bool] = Query(default=None, description="Whether to list intermediate videos."),
    board_id: Optional[str] = Query(
        default=None,
        description="The board id to filter by. Use 'none' to find videos without a board.",
    ),
    order_dir: SQLiteDirection = Query(default=SQLiteDirection.Descending, description="The order of sort"),
    starred_first: bool = Query(default=True, description="Whether to sort by starred videos first"),
    search_term: Optional[str] = Query(default=None, description="The term to search for"),
) -> VideoIdsResult:
    """Gets ordered list of video ids with metadata for optimistic updates"""

    raise HTTPException(status_code=501, detail="Not implemented")


@videos_router.post(
    "/videos_by_ids",
    operation_id="get_videos_by_ids",
    responses={200: {"model": list[VideoDTO]}},
)
async def get_videos_by_ids(
    video_ids: list[str] = Body(embed=True, description="Object containing list of video ids to fetch DTOs for"),
) -> list[VideoDTO]:
    """Gets video DTOs for the specified video ids. Maintains order of input ids."""

    raise HTTPException(status_code=501, detail="Not implemented")
